import {
  CanActivate,
  ExecutionContext,
  ForbiddenException,
  Injectable,
  UnauthorizedException,
} from '@nestjs/common';
import { Request } from 'express';
import { AuthService } from './auth.service';
import { DatabaseService } from 'src/database/database.service';
import { Reflector } from '@nestjs/core';

export class AuthenticationData {
  user: string;
  god: boolean;
}

@Injectable()
export class AuthGuard implements CanActivate {
  constructor(
    private auth: AuthService,
    private db: DatabaseService,
    private reflector: Reflector,
  ) {}

  async canActivate(context: ExecutionContext): Promise<boolean> {
    const request: Request = context.switchToHttp().getRequest();

    const token = this.extractToken(request);
    if (!token) throw new UnauthorizedException();

    const details = await this.auth.getToken(token);
    (request as any)['auth_data'] = {
      user: details.user,
      god: details.god,
    } as AuthenticationData;

    if (details.god) return true;

    const accessLevel = this.reflector.get<string | undefined>(
      'access_level',
      context.getHandler(),
    );

    if (request.params['server']) {
      const memberDetails = await this.db.fetchMemberInfo(
        details.user,
        request.params['server'],
      );

      if (accessLevel) {
        if (!memberDetails.isMember) return false;
        if (accessLevel == 'moderator') {
          if (!memberDetails.isModerator) return false;
        }
        if (accessLevel == 'administrator') {
          if (!memberDetails.isAdmin) return false;
        }
      }
    } else if (accessLevel) return false;
    return true;
  }

  private extractToken(request: Request): string {
    const header = request.header('Authorization');
    if (!header)
      throw new UnauthorizedException('Authorization header missing');
    const [type, token] = header.split(' ');
    if (type != 'Bearer' || !token)
      throw new UnauthorizedException('Invalid or missing token');
    return token;
  }
}

export class NoAuthGuard implements CanActivate {
  async canActivate(context: ExecutionContext): Promise<boolean> {
    const request: Request = context.switchToHttp().getRequest();
    if (request.header('authorization')) {
      throw new ForbiddenException(
        'This route may not be called with authentication',
      );
    }
    return true;
  }
}
