import {
  BadRequestException,
  Injectable,
  InternalServerErrorException,
  UnauthorizedException,
} from '@nestjs/common';
import { DBAuthToken, RE_ULID, generateAuthToken, generateOTP } from 'lib';
import { ulid } from 'ulid';
import { lookup as ipLookup } from 'geoip-lite';
import { DatabaseService } from 'src/database/database.service';
import { LoginResponse } from './auth.controller';
import { UserAuditLogType } from 'lib/dist/cjs/types/database/user_audit_log';

@Injectable()
export class AuthService {
  constructor(private db: DatabaseService) {}

  async getToken(token: string): Promise<DBAuthToken> {
    const tokenData = await this.db.getDb().authTokens.findOne({
      token: token,
    });

    if (!tokenData) throw new UnauthorizedException('Invalid token');
    if (tokenData.expires && tokenData.expires.getTime() <= Date.now()) {
      throw new UnauthorizedException('Token expired');
    }
    if (tokenData.pending) {
      throw new UnauthorizedException('Token not activated yet');
    }
    return tokenData;
  }

  async createLoginRequest(user: string, ip: string): Promise<LoginResponse> {
    if (!new RegExp(RE_ULID).test(user)) {
      throw new BadRequestException(
        "The provided user doesn't appear to be a valid user ID",
      );
    }

    const tokenId = ulid();
    const token = generateAuthToken();
    let otp: string | undefined = undefined;

    // Eliminate possibility of duplicate OTPs
    for (let i = 0; i < 10; i++) {
      const generated = generateOTP();
      const count = await this.db.getDb().authTokens.countDocuments({
        otp: generated,
        user: user,
      });

      if (!count) {
        otp = generated;
        break;
      }
    }

    if (!otp) {
      throw new InternalServerErrorException('Unable to find a free OTP');
    }

    await this.db.getDb().authTokens.insertOne({
      _id: tokenId,
      user: user,
      token: token,
      otp: otp,
      expires: new Date(Date.now() + 1000 * 60 * 5),
    });

    const ipData = ipLookup(ip);

    await this.db.getDb().userAuditLog.insertOne({
      _id: ulid(),
      user: user,
      type: UserAuditLogType.LoginAttempt,
      location: {
        country: ipData?.country,
        region: ipData?.region,
        city: ipData?.city,
      },
    });

    return {
      token: token,
      otp: otp,
    };
  }
}
