import {
  applyDecorators,
  createParamDecorator,
  ExecutionContext,
  InternalServerErrorException,
  SetMetadata,
  UseGuards,
} from '@nestjs/common';
import { AuthenticationData, AuthGuard, NoAuthGuard } from './auth.guard';
import {
  ApiBearerAuth,
  ApiForbiddenResponse,
  ApiUnauthorizedResponse,
} from '@nestjs/swagger';

export const AuthData = createParamDecorator(
  (data: unknown, ctx: ExecutionContext) => {
    const request = ctx.switchToHttp().getRequest();
    if (!request.auth_data)
      throw new InternalServerErrorException(
        'Authentication data not present in request object',
      );
    return request.auth_data as AuthenticationData;
  },
);

export function Auth() {
  return applyDecorators(
    UseGuards(AuthGuard),
    ApiBearerAuth(),
    ApiUnauthorizedResponse({
      description: 'No valid authentication presented',
    }),
    ApiForbiddenResponse({
      description: 'The user does not have access to the resource',
    }),
  );
}

/**
 * Specify which server authentication level is required to access the route.
 * By default any user can access the route, even if they're not a server member.
 * Uses the `server` route parameter.
 */
export function ServerAuth(level: 'member' | 'moderator' | 'administrator') {
  return applyDecorators(SetMetadata('access_level', level));
}

export function NoAuthentication() {
  return applyDecorators(
    UseGuards(NoAuthGuard),
    ApiForbiddenResponse({
      description: 'Access to the resource is forbidden',
    }),
  );
}
