import { HttpException, HttpStatus, Injectable } from '@nestjs/common';
import {
  AutomodDatabase,
  AutomodRedis,
  CreateInfractionReq,
  CreateInfractionRes,
  GetMemberReq,
  GetMemberRes,
  GetMutualServersReq,
  GetMutualServersRes,
  GetServerDetailsReq,
  GetServerDetailsRes,
  connectDb,
  connectRedis,
  createRedisIPCSender,
} from 'lib';

let Db: AutomodDatabase | null = null;
let Redis: AutomodRedis | null = null;
let RcvRedis: AutomodRedis | null = null;
let Messengers: Awaited<ReturnType<typeof createMessengers>> | null = null;

async function createMessengers(redis: AutomodRedis, rcvRedis: AutomodRedis) {
  return {
    getMemberInfo: createRedisIPCSender<GetMemberReq, GetMemberRes>(
      'getMemberInfo',
      redis,
      rcvRedis,
    ),
    createInfraction: createRedisIPCSender<
      CreateInfractionReq,
      CreateInfractionRes
    >('createInfraction', redis, rcvRedis),
    getMutualServers: createRedisIPCSender<
      GetMutualServersReq,
      GetMutualServersRes
    >('getMutualServers', redis, rcvRedis),
    getServerDetails: createRedisIPCSender<
      GetServerDetailsReq,
      GetServerDetailsRes
    >('getServerDetails', redis, rcvRedis),
  };
}

setTimeout(() => {
  // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
  connectDb(process.env.MONGODB_URI!, process.env.MONGODB_DATABASE!).then(
    (db) => (Db = db),
  );

  // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
  connectRedis(process.env.REDIS_URI!).then(async (redis) => {
    Redis = redis;
    if (RcvRedis) Messengers = await createMessengers(Redis, RcvRedis);
  });

  // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
  connectRedis(process.env.REDIS_URI!).then(async (redis) => {
    RcvRedis = redis;
    if (Redis) Messengers = await createMessengers(Redis, RcvRedis);
  });
}, 0);

@Injectable()
export class DatabaseService {
  getDb(): AutomodDatabase {
    if (Db) return Db;
    throw new HttpException(
      'Database not yet available',
      HttpStatus.INTERNAL_SERVER_ERROR,
    );
  }

  getRedis(): AutomodRedis {
    if (Redis) return Redis;
    throw new HttpException(
      'Redis not yet available',
      HttpStatus.INTERNAL_SERVER_ERROR,
    );
  }

  getMessengers(): Awaited<ReturnType<typeof createMessengers>> {
    if (Messengers) return Messengers;
    throw new HttpException(
      'Redis not yet available',
      HttpStatus.INTERNAL_SERVER_ERROR,
    );
  }

  async fetchMemberInfo(user: string, server: string): Promise<GetMemberRes> {
    const key = `members.${server}.${user}`;
    const redis = this.getRedis();
    const data = await redis.GET(key);
    if (data) return JSON.parse(data);

    const newData = await this.getMessengers().getMemberInfo({ server, user });
    await redis.SETEX(key, 5, JSON.stringify(newData));
    return newData;
  }
}
