import {
  BadRequestException,
  Body,
  Controller,
  ForbiddenException,
  Get,
  HttpException,
  HttpStatus,
  Param,
  Post,
  Query,
} from '@nestjs/common';
import {
  IsNumberString,
  IsOptional,
  IsString,
  MinLength,
} from 'class-validator';
import { DBInfraction } from 'lib';
import { InfractionsService } from './infractions.service';
import { ApiBody, ApiOperation, ApiProperty, ApiTags } from '@nestjs/swagger';
import { AuthenticationData } from 'src/auth/auth.guard';
import { Auth, AuthData, ServerAuth } from 'src/auth/authdata.decorator';
import { DatabaseService } from 'src/database/database.service';

export class ListInfractionsData {
  @ApiProperty({ required: false })
  @IsString()
  @IsOptional()
  user?: string;

  @ApiProperty({ required: false })
  @IsString()
  @IsOptional()
  mod?: string;

  @ApiProperty({ required: false })
  @IsNumberString()
  @IsOptional()
  limit?: string;

  @ApiProperty({
    required: false,
    description: 'Unix millis. Returns values before this timestamp.',
    type: 'number',
  })
  @IsString()
  @IsOptional()
  before?: string;

  @ApiProperty({
    required: false,
    description: 'Unix millis. Returns values after this timestamp.',
    type: 'number',
  })
  @IsString()
  @IsOptional()
  after?: string;
}

export class CreateInfractionData {
  @ApiProperty()
  user: string;

  @ApiProperty()
  @IsString()
  @MinLength(1)
  reason: string;
}

@Auth()
@ServerAuth('member')
@ApiTags('infractions')
@Controller('/servers/:server/infractions')
export class InfractionsController {
  constructor(
    private infractions: InfractionsService,
    private db: DatabaseService,
  ) {}

  @Get()
  @ApiOperation({ summary: 'Fetch a list of infractions' })
  async listInfractions(
    @Param('server') server: string,
    @Query() data: ListInfractionsData,
    @AuthData() auth: AuthenticationData,
  ) {
    const limit = Number(data.limit);

    if (data.limit && (limit > 100 || limit < 1)) {
      throw new BadRequestException('Limit must be between 1 and 100');
    }

    if (data.before != null && data.after != null) {
      throw new BadRequestException("Cannot provide both 'before' and 'after'");
    }

    const accessLevel = await this.infractions.getServerAccessLevel(server);
    const member = await this.db.fetchMemberInfo(auth.user, server);

    if (!member.isMember) {
      throw new ForbiddenException('Not a member of this server');
    }
    if (accessLevel != 'all' && !this.infractions.isModerator(member)) {
      if (accessLevel == 'none')
        throw new ForbiddenException(
          'Cannot access infractions in this server',
        );
      if (data.user != auth.user)
        throw new ForbiddenException(
          "No permission to list all infractions in this server. 'user' parameter must be own ID.",
        );
    }

    return await this.infractions.list(server, {
      user: data.user,
      mod: data.mod,
      limit: limit,
      before: data.before ? Number(data.before) : undefined,
      after: data.after ? Number(data.after) : undefined,
    });
  }

  @Get('/:id')
  @ApiOperation({ summary: 'Fetch a single infraction' })
  async getInfraction(
    @Param('server') server: string,
    @Param('id') id: string,
    @AuthData() auth: AuthenticationData,
  ): Promise<DBInfraction> {
    const infraction = await this.infractions.get(server, id);
    if (!infraction) throw new HttpException('Not found', HttpStatus.NOT_FOUND);

    if (
      !auth.god &&
      !(await this.infractions.canAccessInfraction(
        auth.user,
        server,
        infraction,
      ))
    )
      throw new ForbiddenException("You don't have access to this infraction");

    return infraction;
  }

  @Post()
  @ApiOperation({ summary: 'Create an infraction' })
  @ApiBody({ type: CreateInfractionData })
  async createInfraction(
    @Param('server') server: string,
    @Body() data: CreateInfractionData,
    @AuthData() auth: AuthenticationData,
  ) {
    return await this.infractions.create(server, auth.user, data);
  }
}
