import {
  BadRequestException,
  ForbiddenException,
  Injectable,
} from '@nestjs/common';
import { DBInfraction, GetMemberRes } from 'lib';
import type { Filter } from 'mongodb';
import { DatabaseService } from 'src/database/database.service';
import { CreateInfractionData } from './infractions.controller';

@Injectable()
export class InfractionsService {
  constructor(private readonly db: DatabaseService) {}

  async list(
    server: string,
    data: {
      user?: string;
      mod?: string;
      limit?: number;
      before?: number;
      after?: number;
    },
  ) {
    const query: Filter<DBInfraction> = {};

    if (data.user) query.user = data.user;
    if (data.mod) query.mod = data.mod;
    if (data.before) query.date = { $lt: new Date(data.before) };
    if (data.after) query.date = { $gt: new Date(data.after) };

    const result = await this.db
      .getDb()
      .infractions.find(query)
      .limit(data.limit || 50)
      .toArray();

    const count = await this.db.getDb().infractions.countDocuments(query);

    return { count, result };
  }

  async get(server: string, id: string): Promise<DBInfraction | null> {
    return await this.db
      .getDb()
      .infractions.findOne({ '_id.server': server, '_id.infraction': id });
  }

  async create(server: string, mod: string, data: CreateInfractionData) {
    try {
      return await this.db.getMessengers().createInfraction({
        target: data.user,
        reason: data.reason,
        mod: mod,
        server: server,
      });
    } catch (e) {
      console.log(e);
      switch (e) {
        case 'Unauthorised':
          throw new ForbiddenException();
        default:
          throw new BadRequestException(e);
      }
    }
  }

  async getServerAccessLevel(server: string): Promise<'all' | 'own' | 'none'> {
    const serverConfig = await this.db.getDb().servers.findOne({ _id: server });
    return serverConfig?.infraction_visibility || 'own';
  }

  isModerator(member: GetMemberRes) {
    if (!member.isMember) return false;
    return member.permissions.includes('KickMembers');
  }

  async canAccessInfraction(
    user: string,
    server: string,
    infraction: DBInfraction,
  ): Promise<boolean> {
    const accessLevel = await this.getServerAccessLevel(server);
    const memberInfo = await this.db.fetchMemberInfo(user, server);

    if (!memberInfo.isMember) return false;

    if (accessLevel == 'all') return true;
    if (!memberInfo.permissions.includes('KickMembers')) return false;
    if (accessLevel == 'own' && infraction.user == memberInfo.id) return true;
    return false;
  }
}
