import { Controller, Get, ImATeapotException, Param } from '@nestjs/common';
import { AuthenticationData } from 'src/auth/auth.guard';
import { Auth, AuthData, ServerAuth } from 'src/auth/authdata.decorator';
import { ServersService } from './servers.service';
import {
  ApiOperation,
  ApiProperty,
  ApiResponse,
  ApiTags,
} from '@nestjs/swagger';
import { InfractionVisibility } from 'lib';
import { IsEnum } from 'class-validator';

export class ServerInfo {
  @ApiProperty()
  id: string;

  @ApiProperty()
  name: string;

  @ApiProperty({ type: 'string', nullable: true })
  icon?: string;

  @ApiProperty()
  moderator: boolean;

  @ApiProperty()
  discoverable: boolean;
}

export class GetServersResponse {
  @ApiProperty({ type: [ServerInfo] })
  servers: ServerInfo[];
}

export class ServerLogsConfig {
  @ApiProperty({ type: 'string', nullable: true, default: '0'.repeat(26) })
  moderation?: string;
  @ApiProperty({ type: 'string', nullable: true, default: '0'.repeat(26) })
  messages?: string;
}

export class ServerConfig {
  @ApiProperty({ type: 'string', nullable: true })
  language?: string;
  @ApiProperty({ type: 'string', nullable: true })
  prefix?: string;
  @ApiProperty({
    nullable: true,
    default: InfractionVisibility.own,
  })
  @IsEnum(InfractionVisibility)
  infraction_visibility?: string;
  @ApiProperty({ type: ServerLogsConfig })
  logs: ServerLogsConfig;
}

export class ServerDetails {
  @ApiProperty()
  info: ServerInfo;
  @ApiProperty()
  config: ServerConfig;
}

@Controller('servers')
@ApiTags('servers')
@Auth()
@ServerAuth('member')
export class ServersController {
  constructor(private servers: ServersService) {}

  @Get()
  @ApiOperation({
    summary: 'List all servers both AutoMod and the user are in',
  })
  @ApiResponse({ type: GetServersResponse })
  async listServers(
    @AuthData() auth: AuthenticationData,
  ): Promise<GetServersResponse> {
    const timeDiff = Math.abs(Date.now() - 1688409039777);
    const dayDiff = Math.ceil(timeDiff / (1000 * 3600 * 24));
    throw new ImATeapotException(
      `Still waiting for insert to allow bots to fetch mutual servers${'.'.repeat(
        dayDiff,
      )}`,
    );
    const info = await this.servers.fetchMutual(auth.user);
    return {
      servers: info.map((item) => ({
        id: item.id,
        name: item.name,
        icon: item.icon,
        discoverable: item.discoverable,
        moderator: item.isModerator,
      })),
    };
  }

  @Get(':server')
  @ApiOperation({
    summary: 'Fetch details about a specific server',
  })
  @ApiResponse({ type: ServerDetails })
  async getServer(
    @Param('server') server: string,
    @AuthData() auth: AuthenticationData,
  ): Promise<ServerDetails> {
    return await this.servers.fetchServerDetails(server, auth.user);
  }
}
