/* eslint-disable @typescript-eslint/no-non-null-assertion */
import {
    Command,
    RE_ULID,
    canModifyServerSettings,
    embedMsg,
    flattenObject,
} from "lib";
import { langs, t } from "../../modules/i18n.js";

type SettingsOption = {
    options: string[] | ((value: string) => boolean); // Acceptable options
    validationHint?: string; // Displayed if options is a function
    dbKey: string; // The key on the user database entry
    type: "user" | "server";
    displayValue?: (value: string) => string; // Changes how the value is displayed
    description: string; // english only for now
};

const SETTINGS: { [key: string]: SettingsOption } = {
    language: {
        options: Object.keys(langs),
        dbKey: "language",
        type: "user",
        displayValue: (value) => {
            const lang = langs[value];
            if (!lang) return value;
            return `${lang.emoji} ${lang.nameLocalised} (${value})`;
        },
        description: "The language AutoMod will use to interact with you",
    },
    pronouns: {
        options: ["they", "she", "he"],
        dbKey: "pronouns",
        type: "user",
        description: "Choose how AutoMod should refer to you",
    },
    prefix: {
        options: (value) => value.length < 10,
        validationHint: "length < 10",
        dbKey: "prefix",
        type: "server",
        description: "What prefix AutoMod should respond to in this server",
    },
    serverlang: {
        options: Object.keys(langs),
        dbKey: "language",
        type: "server",
        displayValue: (value) => {
            const lang = langs[value];
            if (!lang) return value;
            return `${lang.emoji} ${lang.nameLocalised} (${value})`;
        },
        description: "The language used in log messages sent to this server",
    },
    modlog: {
        options: (value) => new RegExp(RE_ULID).test(value),
        validationHint: "<Channel ID>",
        dbKey: "logs.moderation",
        type: "server",
        description: "Select a channel ID to send moderation notifications to",
    },
    messagelog: {
        options: (value) => new RegExp(RE_ULID).test(value),
        validationHint: "<Channel ID>",
        dbKey: "logs.messages",
        type: "server",
        description: "Placeholder setting, currently unused",
    },
    infraction_visibility: {
        options: ["all", "own", "none"],
        dbKey: "infraction_visibility",
        type: "server",
        displayValue(value) {
            switch (value) {
                case "all":
                    return "Server members can view all infractions";
                case "none":
                    return "Server members can't view any infractions";
                default:
                    return "Server members can view their own infractions";
            }
        },
        description: "Select whose infractions server members can view",
    },
};

function validateValue(setting: SettingsOption, value: string | null): boolean {
    if (value == null) return true;
    if (typeof setting.options == "function") {
        return setting.options(value);
    }
    return setting.options.includes(value);
}

function printOptions(setting: SettingsOption): string {
    if (typeof setting.options == "function") {
        return setting.validationHint ?? "any";
    }
    return setting.options.map((option) => `\`${option}\``).join(", ");
}

export default {
    name: "config",
    aliases: ["setting", "settings"],
    description: "Personal bot configuration",
    async run(message, args, client, context) {
        const setting = args.shift();
        const currentUserConfig = await client.db.users.findOne({
            _id: message.authorId,
        });

        if (!setting) {
            let text = Object.entries(SETTINGS)
                .filter(
                    ([_key, value]) => message.server || value.type == "user",
                )
                .map(([key, value]) => {
                    const val = flattenObject(
                        value.type == "user"
                            ? currentUserConfig
                            : context.serverConfig,
                    )?.[value.dbKey as never];
                    const displayVal =
                        value.displayValue && val
                            ? value.displayValue(val)
                            : val;

                    return `**${key}** (${value.type}): \`${
                        displayVal || "unset"
                    }\``;
                })
                .join("\n");

            text +=
                "\n> " +
                (await t("commands.config.usage_hint", message.author, {
                    command: `${context.serverPrefix}${context.commandName}`,
                }));

            await message.reply(
                embedMsg(
                    text,
                    await t("commands.config.title", message.author),
                ),
            );
        } else if (!args.length) {
            const settingData = SETTINGS[setting];
            if (!settingData) {
                return await message.reply(
                    embedMsg(
                        await t(
                            "commands.config.unknown_setting",
                            message.author,
                            {
                                setting,
                            },
                        ),
                        await t("commands.config.title", message.author),
                        "ERROR",
                    ),
                );
            } else {
                const relevantConfig = flattenObject(
                    settingData.type == "user"
                        ? currentUserConfig
                        : context.serverConfig,
                );
                const currentValue =
                    relevantConfig?.[settingData.dbKey as never];

                return await message.reply(
                    embedMsg(
                        "### " +
                            (await t(
                                "commands.config.setting_info.title",
                                message.author,
                                {
                                    setting,
                                },
                            )) +
                            "\n" +
                            (await t(
                                "commands.config.setting_info.type",
                                message.author,
                                {
                                    type: `\`${settingData.type}\``,
                                },
                            )) +
                            "\n" +
                            (await t(
                                "commands.config.setting_info.current",
                                message.author,
                                {
                                    value:
                                        (settingData.displayValue &&
                                        currentValue
                                            ? `\`${settingData.displayValue(
                                                  currentValue,
                                              )}\``
                                            : currentValue) || "-",
                                },
                            )) +
                            "\n" +
                            (await t(
                                "commands.config.setting_info.options",
                                message.author,
                                {
                                    options: printOptions(settingData),
                                },
                            )),
                        await t("commands.config.title", message.author),
                    ),
                );
            }
        } else {
            const settingData = SETTINGS[setting];
            if (!settingData) {
                await message.reply(
                    embedMsg(
                        await t(
                            "commands.config.unknown_setting",
                            message.author,
                            {
                                setting,
                            },
                        ),
                        await t("commands.config.title", message.author),
                        "ERROR",
                    ),
                );
                return;
            }
            const relevantConfig = flattenObject(
                settingData.type == "user"
                    ? currentUserConfig
                    : context.serverConfig,
            );

            if (settingData.type == "server") {
                if (!message.server || !message.member)
                    return await message.reply(
                        embedMsg(
                            await t(
                                "commands.config.needs_server",
                                message.author,
                            ),
                            await t("commands.config.title", message.author),
                            "ERROR",
                        ),
                    );
                else if (!canModifyServerSettings(message.member)) {
                    return await message.reply(
                        embedMsg(
                            await t(
                                "commands.config.no_permission",
                                message.author,
                            ),
                            await t("commands.config.title", message.author),
                            "ERROR",
                        ),
                    );
                }
            }

            const currentVal = relevantConfig?.[settingData.dbKey as never];
            const newVal = ["clear", "reset"].includes(args[0])
                ? null
                : args.join(" ");

            if (!validateValue(settingData, newVal)) {
                return await message.reply(
                    embedMsg(
                        await t(
                            "commands.config.invalid_value",
                            message.author,
                            {
                                values: printOptions(settingData),
                            },
                        ),
                        await t("commands.config.title", message.author),
                        "ERROR",
                    ),
                );
            }

            if (currentVal == newVal) {
                return await message.reply(
                    embedMsg(
                        await t(
                            "commands.config.setting_unchanged",
                            message.author,
                            {
                                value: `\`${currentVal}\``,
                                setting: `\`${setting}\``,
                            },
                        ),
                        await t("commands.config.title", message.author),
                    ),
                );
            }

            if (settingData.type == "user") {
                await client.db.users.updateOne(
                    { _id: message.authorId },
                    { $set: { [settingData.dbKey]: newVal } },
                    { upsert: true },
                );
            } else if (settingData.type == "server") {
                await client.db.servers.updateOne(
                    { _id: message.server!.id },
                    { $set: { [settingData.dbKey]: newVal } },
                    { upsert: true },
                );
            }

            await message.reply(
                embedMsg(
                    await t("commands.config.setting_updated", message.author, {
                        old: `\`${currentVal || "-"}\``,
                        value: `\`${newVal || "-"}\``,
                        setting: `\`${setting}\``,
                    }),
                    await t("commands.config.title", message.author),
                    "SUCCESS",
                ),
            );
        }
    },
} as Command;
