import {
    Command,
    DBInfraction,
    InfractionType,
    TimedActionState,
    TimedActionType,
    createInfraction,
    embed,
    generateInfractionId,
    separateUser,
    splitTimeInput,
} from "lib";
import { AxiosError } from "axios";
import { t } from "../../modules/i18n.js";
import { decodeTime, ulid } from "ulid";
import { API, User } from "revolt.js";
import {
    checkConditions,
    checkRemovalEligibility,
    sendInfractionLog,
} from "../../util/infractions.js";
import { cancelExecution, scheduleEvent } from "../../modules/timedActions.js";
import { logger } from "../../index.js";

/**
 * todo: listen for unban events and remove tempbans from database if the user is unbanned manually
 */

export default {
    name: "ban",
    aliases: ["banish", "kill"],
    description: "Ban a member from the server, temporarily or permanently",
    permissions: ["BanMembers"],
    serverOnly: true,
    async run(message, args, client, context) {
        if (!message.authorId) throw new Error("authorId is not set");
        if (!message.server?.id) throw new Error("server.id is not set");
        if (!message.member) throw new Error("member is not set");
        if (!message.server.member) throw new Error("server.member is not set");

        const { reason, users, members, certain } = await separateUser(
            message,
            args,
            client,
            true,
        );
        const { text, time } = splitTimeInput(reason?.split(" ") || []);

        if (
            !(await checkConditions(
                message,
                client,
                users,
                0,
                // We also want to show a confirmation prompt in case someone types /kill as a joke
                certain && context.commandName != "kill",
            ))
        ) {
            return;
        }

        const embeds: API.SendableEmbed[] = [];
        for (const user of users) {
            try {
                if (user instanceof User) {
                    const member = members?.find((m) => m.id.user == user.id);
                    if (member) {
                        // Potential privilege escalation? in theory `separateUser()` always explicitly
                        // fetches the member object, but if that's for some reason not the case, users
                        // will be able to ban someone ranked higher than them.
                        const error = await checkRemovalEligibility(
                            message,
                            member,
                            user,
                            client,
                            false,
                        );

                        if (error) {
                            embeds.push(error);
                            continue;
                        }
                    }
                }

                const banExpires = time ? new Date(Date.now() + time) : null;
                const infraction: DBInfraction = {
                    _id: {
                        infraction: await generateInfractionId(
                            message.server,
                            client.db,
                        ),
                        server: message.server.id,
                    },
                    date: new Date(decodeTime(message.id)),
                    reason: text || null,
                    user: user.id,
                    moderator: message.authorId,
                    context: {
                        message: message.id,
                        channel: message.channelId,
                        server: message.channel?.serverId,
                    },
                    ...(banExpires
                        ? {
                              type: InfractionType.BanTemporary,
                              expires: banExpires,
                          }
                        : { type: InfractionType.BanPermanent }),
                };

                try {
                    await message.server.banUser(user.id, {
                        // TODO: Ban reason will not be translated for now since there is no server wide language option yet
                        reason: `${
                            time ? "Temporarily" : "Permanently"
                        } banned by ${message.author?.username}#${
                            message.author?.discriminator
                        } - Reason: ${reason || "None"} - Infraction ID: ${
                            infraction._id.infraction
                        }`,
                    });
                } catch (e) {
                    switch ((e as AxiosError).response?.status) {
                        case 500: {
                            logger.warn(
                                "Received error 500 on ban; Assuming this means the user is already banned.",
                            );
                            break;
                        }
                        case 429: {
                            throw new Error(
                                "AutoMod has hit a rate limit and cannot ban more users at this time. Please retry in a couple seconds.",
                            );
                            break;
                        }
                        default:
                            throw e;
                    }
                }

                // If there was a previous pending tempban, we want to cancel it.
                const actions = await client.db.timedActions
                    .find({
                        action: TimedActionType.Unban,
                        state: TimedActionState.Pending,
                        server: message.server.id,
                        user: user.id,
                    })
                    .toArray();

                for (const action of actions) {
                    await cancelExecution(action._id);
                }

                const count = await createInfraction(
                    infraction,
                    message.server,
                    client,
                );

                if (banExpires) {
                    await scheduleEvent({
                        _id: ulid(),
                        action: TimedActionType.Unban,
                        state: TimedActionState.Pending,
                        date: banExpires,
                        infraction: infraction._id,
                    });
                }

                embeds.push(
                    embed(
                        await t("commands.ban.user_banned", message.author, {
                            id: infraction._id.infraction,
                            user: infraction.user,
                            reason: infraction.reason
                                ? `\`${infraction.reason}\``
                                : "*None*",
                            count,
                            expires: banExpires
                                ? `<t:${Math.round(
                                      banExpires.getTime() / 1000,
                                  )}:R>`
                                : `\`${await t(
                                      "infractions.permanent",
                                      message.author,
                                  )}\``,
                        }),
                        await t(
                            "infractions.infraction_created",
                            message.author,
                        ),
                        "SUCCESS",
                    ),
                );
                await sendInfractionLog(infraction, context.serverConfig);
            } catch (e) {
                embeds.push(
                    embed(
                        `\`\`\`js\n${e}\n\`\`\``,
                        await t("error_generic", message.author),
                        "ERROR",
                    ),
                );
            }
        }

        await message.reply({
            content: await t("infractions.operation_completed", message.author),
            embeds,
        });
    },
} as Command;
