import {
    Command,
    DBServer,
    InfractionType,
    ModActionType,
    ServerLogCategory,
    embed,
    embedMsg,
    fetchUser,
    hideNewlines,
    separateUser,
    yesNoMessage,
} from "lib";
import { joinTranslatedDetails, t, tEmbedMsg } from "../../modules/i18n.js";
import { API, User } from "revolt.js";
import { dispatchLogMessage } from "../../modules/server_logs.js";
import { ulid } from "ulid";

const EMOJIS: { [key in InfractionType]: string } = {
    ban_permanent: "🔨",
    ban_temporary: "🔨🕑",
    timeout: "🔇",
    kick: "🦶",
    warn: "⚠️",
};

// TODO deduplicate some of the permission checks - We don't need to
// copypaste the same 6 line translation function call this often

export default {
    name: "infractions",
    aliases: ["warns", "warnings"],
    description: "List or modify server member infractions",
    serverOnly: true,
    async run(message, args, client, context) {
        const isMod = !!message.member?.hasPermission(
            message.server!,
            "KickMembers",
        );

        if (args[0]?.startsWith("#")) {
            const id = args[0].substring(1);
            const infraction = await client.db.infractions.findOne({
                "_id.infraction": { $in: [id, id.padStart(4, "0")] }, // Retrieve the infraction even if the user types the ID without leading zeroes
                "_id.server": message.server!.id,
            });

            if (!infraction) {
                await message.reply(
                    await tEmbedMsg(
                        message.author,
                        "commands.infractions.infraction_does_not_exist",
                        "commands.infractions.title",
                        "ERROR",
                        { id },
                    ),
                );
                return;
            }

            const user = await fetchUser(infraction.user, client);
            const mod =
                infraction.moderator == "System"
                    ? undefined
                    : await fetchUser(infraction.moderator, client);
            const messageUrl = infraction.context
                ? infraction.context.server
                    ? `https://app.revolt.chat/server/${infraction.context.server}/channel/${infraction.context.channel}/${infraction.context.message}`
                    : `https://app.revolt.chat/channel/${infraction.context.channel}/${infraction.context.message}`
                : undefined;

            const details = await joinTranslatedDetails(
                "commands.infractions.details",
                message.author,
                context.serverConfig || undefined,
                [
                    {
                        key: "target",
                        item: "username",
                        username: user
                            ? `${user.username}#${user.discriminator} (\`${user.id}\`)`
                            : infraction.user,
                    },
                    {
                        key: "moderator",
                        item: "username",
                        username:
                            infraction.moderator == "System"
                                ? "System"
                                : mod
                                ? `${mod.username}#${mod.discriminator} (\`${mod.id}\`)`
                                : infraction.moderator,
                    },
                    {
                        key: "created",
                        item: "ts",
                        ts: `<t:${Math.round(
                            infraction.date.getTime() / 1000,
                        )}:d>`,
                        tsRelative: `<t:${Math.round(
                            infraction.date.getTime() / 1000,
                        )}:R>`,
                    },
                    {
                        key: "type",
                        item: "type",
                        type:
                            EMOJIS[infraction.type] +
                            " " +
                            (await t(
                                "infractions.names." + infraction.type,
                                message.author,
                            )),
                    },
                    {
                        key: "reason",
                        item: "reason",
                        reason: hideNewlines(
                            infraction.reason || "*No reason provided*",
                        ),
                    },
                    {
                        key: "context",
                        item: "url",
                        url: messageUrl,
                    },
                    {
                        key: "revoked",
                        item: "revoked",
                        revoked: infraction.revoked,
                    },
                ],
            );

            switch (args[1]?.toLowerCase()) {
                case undefined:
                case "info":
                case "show": {
                    // Permission checks
                    if (!isMod) {
                        if (
                            context.serverConfig?.infraction_visibility ==
                            "none"
                        ) {
                            await message.reply(
                                embedMsg(
                                    await t(
                                        "commands.infractions.members_cant_view_infractions",
                                        message.author,
                                    ),
                                    await t(
                                        "commands.infractions.title",
                                        message.author,
                                    ),
                                    "ERROR",
                                ),
                            );
                            return;
                        } else if (
                            context.serverConfig?.infraction_visibility !=
                                "all" &&
                            infraction.user != message.authorId
                        ) {
                            await message.reply(
                                embedMsg(
                                    await t(
                                        "commands.infractions.members_can_only_view_own",
                                        message.author,
                                    ),
                                    await t(
                                        "commands.infractions.title",
                                        message.author,
                                    ),
                                    "ERROR",
                                ),
                            );
                            return;
                        }
                    }

                    // The actual command
                    const text =
                        (await t(
                            "commands.infractions.details.header",
                            message.author,
                            { id: infraction._id.infraction },
                        )) +
                        "\n" +
                        details;

                    await message.reply(
                        embedMsg(
                            text,
                            await t(
                                "commands.infractions.title",
                                message.author,
                            ),
                            "INFO",
                        ),
                    );
                    break;
                }
                case "revoke":
                case "delete":
                case "remove":
                case "del":
                case "rm": {
                    if (!isMod) {
                        await message.reply(
                            embedMsg(
                                await t(
                                    "commands.infractions.no_permission",
                                    message.author,
                                ),
                                await t(
                                    "commands.infractions.title",
                                    message.author,
                                ),
                                "ERROR",
                            ),
                        );
                        return;
                    }

                    if (infraction.revoked) {
                        return await message.reply(
                            await tEmbedMsg(
                                message.author,
                                "commands.infractions.already_revoked",
                                "commands.infractions.title",
                                "ERROR",
                            ),
                        );
                    }

                    const result = await yesNoMessage(
                        message.channel!,
                        client,
                        message.authorId!,
                        `#### ${await t(
                            "commands.infractions.revoke_confirm",
                            message.author,
                            { id: infraction._id.infraction },
                        )}\n${details}`,
                        await t("commands.infractions.title", message.author),
                        await t(
                            "commands.infractions.revoke_confirmed",
                            message.author,
                            { id: infraction._id.infraction },
                        ),
                        await t(
                            "commands.infractions.revoke_cancelled",
                            message.author,
                        ),
                    );
                    if (result) {
                        await client.db.infractions.updateOne(
                            { _id: infraction._id },
                            { $set: { revoked: true } },
                        );
                    }
                    await dispatchLogMessage(
                        {
                            _id: ulid(),
                            category: ServerLogCategory.ModAction,
                            type: ModActionType.RemoveInfraction,
                            infraction: infraction._id,
                            user: message.authorId!,
                            server: infraction._id.server,
                        },
                        client,
                        context.serverConfig,
                    );
                    break;
                }
                case "update":
                case "modify":
                case "change":
                case "edit": {
                    if (!isMod) {
                        await message.reply(
                            embedMsg(
                                await t(
                                    "commands.infractions.no_permission",
                                    message.author,
                                ),
                                await t(
                                    "commands.infractions.title",
                                    message.author,
                                ),
                                "ERROR",
                            ),
                        );
                        return;
                    }

                    if (infraction.revoked) {
                        return await message.reply(
                            await tEmbedMsg(
                                message.author,
                                "commands.infractions.cant_edit_revoked",
                                "commands.infractions.title",
                                "ERROR",
                            ),
                        );
                    }

                    const newReason = args.slice(2).join(" ");

                    if (!newReason)
                        return await message.reply(
                            await tEmbedMsg(
                                message.author,
                                "commands.infractions.no_new_reason",
                                "commands.infractions.title",
                                "ERROR",
                                { command: args[1].toLocaleLowerCase() },
                            ),
                        );

                    const result = await yesNoMessage(
                        message.channel!,
                        client,
                        message.authorId!,
                        `#### ${await t(
                            "commands.infractions.edit_confirm",
                            message.author,
                            { id: infraction._id.infraction },
                        )}\n##### ${await t(
                            "commands.infractions.edit_new_reason",
                            message.author,
                            { reason: hideNewlines(newReason) },
                        )}\n${details}`,
                        await t("commands.infractions.title", message.author),
                        await t(
                            "commands.infractions.edit_confirmed",
                            message.author,
                            { id: infraction._id.infraction },
                        ),
                        await t(
                            "commands.infractions.edit_cancelled",
                            message.author,
                        ),
                    );
                    if (result) {
                        // We filter our query by unrevoked infractions only. This prevents users from
                        // revoking it before they confirm the edit.
                        await client.db.infractions.updateOne(
                            { _id: infraction._id, revoked: { $ne: true } },
                            {
                                $set: { reason: newReason },
                            },
                        );

                        await dispatchLogMessage(
                            {
                                _id: ulid(),
                                category: ServerLogCategory.ModAction,
                                type: ModActionType.EditInfraction,
                                user: message.authorId!,
                                server: infraction._id.server,
                                infraction: infraction._id,
                                oldReason: infraction.reason,
                            },
                            client,
                            context.serverConfig,
                        );
                    }
                    break;
                }
                default:
                    await message.reply(
                        await tEmbedMsg(
                            message.author,
                            "commands.infractions.unknown_action",
                            "commands.infractions.title",
                            "ERROR",
                            { action: args[1] },
                        ),
                    );
            }
            return;
        }

        if (context.serverConfig?.infraction_visibility === "none" && !isMod) {
            return await message.reply(
                embedMsg(
                    await t(
                        "commands.infractions.members_cant_view_infractions",
                        message.author,
                    ),
                    await t("commands.infractions.title", message.author),
                    "ERROR",
                ),
            );
        }

        const { users: targets, notFound } = await separateUser(
            message,
            args,
            client,
            true,
        );

        if (!targets.length && !notFound && message.author) {
            targets.push(message.author);
        }

        if (!targets.length || notFound) {
            await message.reply(
                embedMsg(
                    await t(
                        "commands.infractions.user_not_found",
                        message.author,
                        {
                            count: notFound,
                        },
                    ),
                    await t("commands.infractions.title", message.author),
                    "ERROR",
                ),
            );
            return;
        }

        if (
            !isMod &&
            context.serverConfig?.infraction_visibility !== "all" &&
            targets.filter((u) => u.id != message.authorId).length
        ) {
            return await message.reply(
                embedMsg(
                    await t(
                        "commands.infractions.members_can_only_view_own",
                        message.author,
                    ),
                    await t("commands.infractions.title", message.author),
                    "ERROR",
                ),
            );
        }

        const allInfractions = await client.db.infractions
            .find({
                "_id.server": message.server!.id,
                user: { $in: targets.map((u) => u.id) },
                revoked: { $ne: true },
            })
            .toArray();

        // We want newest on top
        allInfractions.reverse();

        const embeds: API.SendableEmbed[] = [];

        for (const user of targets) {
            const infractions = allInfractions.filter((i) => i.user == user.id);

            let text = await t(
                "commands.infractions.embed_header",
                message.author,
                {
                    count: infractions.length,
                    username:
                        user instanceof User
                            ? `${user.username}#${user.discriminator}`
                            : user.id,
                },
            );

            let i = 0;
            for (const infraction of infractions) {
                let line = `${EMOJIS[infraction.type]} ${hideNewlines(
                    infraction.reason || "*No reason provided*",
                )}`;

                if (infraction.context) {
                    const messageUrl = infraction.context.server
                        ? `https://app.revolt.chat/server/${infraction.context.server}/channel/${infraction.context.channel}/${infraction.context.message}`
                        : `https://app.revolt.chat/channel/${infraction.context.channel}/${infraction.context.message}`;
                    line += ` - [Context](${messageUrl})`;
                }

                // TODO internationalise this
                line +=
                    `\nID: \`${infraction._id.infraction}\`` +
                    ` - <t:${Math.round(infraction.date.getTime() / 1000)}:R> `;
                if (infraction.moderator != "System") {
                    const mod = await fetchUser(infraction.moderator, client);
                    if (mod) {
                        line += `(By ${mod.username}#${mod.discriminator})`;
                    } else {
                        line += `(By \`${infraction.moderator}\`)`;
                    }
                } else {
                    line += "(System)";
                }

                line = "\n- " + line.replaceAll("\n", "\\\n");
                if ((text + line).length > 1600) {
                    text +=
                        "\n" +
                        (await t(
                            "commands.infractions.more_infractions",
                            message.author,
                            {
                                count: infractions.length - i,
                                url: "https://example.org",
                            },
                        ));
                    break;
                }

                text += line;
                i++;
            }

            embeds.push(
                embed(
                    text,
                    await t("commands.infractions.title", message.author),
                    "INFO",
                ),
            );
        }

        // TODO would be smarter to batch multiple embeds in a single message,
        // but Revolt limits us to 2000 characters per message, not per embed.
        // Currently can't be bothered.
        let i = 0;
        for (const currentEmbed of embeds) {
            if (i++ == 0)
                await message.reply({
                    embeds: [
                        embed(
                            await t(
                                "commands.infractions.details_hint",
                                message.author,
                                { prefix: context.serverPrefix },
                            ),
                        ),
                        currentEmbed,
                    ],
                });
            else await message.channel?.sendMessage({ embeds: [currentEmbed] });
        }
    },
} as Command;
