import {
    Command,
    DBInfraction,
    InfractionType,
    createInfraction,
    embed,
    generateInfractionId,
    separateUser,
} from "lib";
import { t } from "../../modules/i18n.js";
import { API } from "revolt.js";
import {
    checkConditions,
    checkRemovalEligibility,
    sendInfractionLog,
} from "../../util/infractions.js";
import { decodeTime } from "ulid";

export default {
    name: "kick",
    aliases: ["eject"],
    description: "Kick a member from the server",
    permissions: ["KickMembers"],
    serverOnly: true,
    async run(message, args, client, context) {
        if (!message.authorId) throw new Error("authorId is not set");
        if (!message.server?.id) throw new Error("server.id is not set");
        if (!message.member) throw new Error("member is not set");
        if (!message.server.member) throw new Error("server.member is not set");

        const { reason, users, members, certain, notFound } =
            await separateUser(message, args, client, false);

        if (
            !(await checkConditions(message, client, users, notFound, certain))
        ) {
            return;
        }

        const embeds: API.SendableEmbed[] = [];
        for (const user of users) {
            try {
                const member = members?.find((m) => m.id.user == user.id);
                if (!member) {
                    embeds.push(
                        embed(
                            await t(
                                "commands.kick.not_a_member",
                                message.author,
                                { userid: user.id },
                            ),
                            await t("infractions.title_create", message.author),
                            "WARN",
                        ),
                    );
                    continue;
                }

                const error = await checkRemovalEligibility(
                    message,
                    member,
                    user,
                    client,
                    false,
                );

                if (error) {
                    embeds.push(error);
                    continue;
                }

                const infraction: DBInfraction = {
                    _id: {
                        infraction: await generateInfractionId(
                            message.server,
                            client.db,
                        ),
                        server: message.server.id,
                    },
                    date: new Date(decodeTime(message.id)),
                    type: InfractionType.Kick,
                    reason,
                    user: user.id,
                    moderator: message.authorId,
                    context: {
                        message: message.id,
                        channel: message.channelId,
                        server: message.channel?.serverId,
                    },
                };

                await member.kick();

                const count = await createInfraction(
                    infraction,
                    message.server,
                    client,
                );
                embeds.push(
                    embed(
                        await t("commands.kick.user_kicked", message.author, {
                            id: infraction._id.infraction,
                            user: infraction.user,
                            reason: infraction.reason
                                ? `\`${infraction.reason}\``
                                : "*None*",
                            count,
                        }),
                        await t(
                            "infractions.infraction_created",
                            message.author,
                        ),
                        "SUCCESS",
                    ),
                );
                await sendInfractionLog(infraction, context.serverConfig);
            } catch (e) {
                embeds.push(
                    embed(
                        `\`\`\`js\n${e}\n\`\`\``,
                        await t("error_generic", message.author),
                    ),
                );
            }
        }

        await message.reply({
            content: await t("infractions.operation_completed", message.author),
            embeds,
        });
    },
} as Command;
