import {
    Command,
    DBInfraction,
    InfractionType,
    createInfraction,
    embed,
    embedMsg,
    generateInfractionId,
    separateUser,
    splitTimeInput,
    yesNoMessage,
} from "lib";
import { t } from "../../modules/i18n.js";
import { decodeTime, ulid } from "ulid";
import { API, User } from "revolt.js";
import {
    checkConditions,
    clearMemberTimeout,
    sendInfractionLog,
    timeoutMember,
} from "../../util/infractions.js";

export default {
    name: "timeout",
    aliases: ["mute", "shush"],
    description: "Time out a member",
    permissions: ["TimeoutMembers"],
    serverOnly: true,
    async run(message, args, client, context) {
        if (!message.authorId) throw new Error("authorId is not set");
        if (!message.server?.id) throw new Error("server.id is not set");
        if (!message.channel?.id) throw new Error("channel.id is not set");

        const { reason, users, certain, notFound } = await separateUser(
            message,
            args,
            client,
            true,
        );
        const { text, time } = splitTimeInput(reason?.split(" ") || []);

        if (
            !(await checkConditions(message, client, users, notFound, certain))
        ) {
            return;
        }

        if (!message.server.havePermission("TimeoutMembers")) {
            return await message.reply(
                embedMsg(
                    await t("commands.timeout.no_permission", message.author),
                    await t("commands.timeout.title", message.author),
                ),
            );
        }

        if (!time) {
            const result = await yesNoMessage(
                message.channel,
                client,
                message.authorId,
                await t("commands.timeout.prompt_clear", message.author, {
                    count: users.length,
                }),
                await t("commands.timeout.title_clear", message.author),
            );
            if (!result) return;
        }

        const embeds: API.SendableEmbed[] = [];
        for (const user of users) {
            try {
                if (time) {
                    const infraction: DBInfraction = {
                        _id: {
                            infraction: await generateInfractionId(
                                message.server,
                                client.db,
                            ),
                            server: message.server.id,
                        },
                        date: new Date(decodeTime(message.id)),
                        type: InfractionType.Timeout,
                        expires: new Date(decodeTime(message.id) + time),
                        reason: text,
                        user: user.id,
                        moderator: message.authorId,
                        context: {
                            message: message.id,
                            channel: message.channelId,
                            server: message.channel?.serverId,
                        },
                    };

                    await timeoutMember(
                        user.id,
                        message.server.id,
                        infraction.expires,
                    );
                    const count = await createInfraction(
                        infraction,
                        message.server,
                        client,
                    );
                    embeds.push(
                        embed(
                            await t(
                                "commands.timeout.user_timed_out",
                                message.author,
                                {
                                    id: infraction._id.infraction,
                                    user: infraction.user,
                                    reason: infraction.reason
                                        ? `\`${infraction.reason}\``
                                        : "*None*",
                                    expires: `<t:${Math.round(
                                        infraction.expires.getTime() / 1000,
                                    )}:R>`,
                                    count,
                                },
                            ),
                            await t(
                                "infractions.infraction_created",
                                message.author,
                            ),
                            "SUCCESS",
                        ),
                    );
                    await sendInfractionLog(infraction, context.serverConfig);
                } else {
                    await clearMemberTimeout(user.id, message.server.id);
                    embeds.push(
                        embed(
                            await t(
                                "commands.timeout.cleared_timeout",
                                message.author,
                                {
                                    username:
                                        user instanceof User
                                            ? `${user.username}#${user.displayName}`
                                            : "*Unknown*",
                                    displayname:
                                        user instanceof User
                                            ? user.displayName
                                            : "*Unknown*",
                                    userid: user.id,
                                },
                            ),
                            await t(
                                "infractions.infraction_created",
                                message.author,
                            ),
                            "SUCCESS",
                        ),
                    );
                }
            } catch (e) {
                embeds.push(
                    embed(
                        `\`\`\`js\n${e}\n\`\`\``,
                        await t("error_generic", message.author),
                    ),
                );
            }
        }

        await message.reply({
            content: await t("infractions.operation_completed", message.author),
            embeds,
        });
    },
} as Command;
