import {
    BotPermission,
    Command,
    CommandContext,
    dirname,
    embedMsg,
    enabledInServer,
    recursiveReaddir,
} from "lib";
import path from "path";
import { client, logger } from "../index.js";
import { t } from "./i18n.js";
import { META } from "../meta.js";
import { Message, ServerMember, User } from "revolt.js";
import { Permission } from "revolt.js/src/permissions/definitions.js";

// TODO don't hardcode these
export const BOT_ADMINS = ["01EXAF3KX65608AJ4NG27YG1HM"];
export const IS_SELF_HOSTED = ![
    "https://api.revolt.chat",
    "https://revolt.chat/api",
].includes(client.options.baseURL);

const DEFAULT_PREFIX = process.env.PREFIX || "/";
const RE_PREFIX_MENTION = new RegExp(`^\\s?<@${client.user?.id}>\\s?`, "g");

logger.info(`Default prefix is ${DEFAULT_PREFIX}`);

// Commands are dynamically imported and blindly expected to have the correct type.
// This sometimes causes issues because TypeScript doesn't automatically delete old files.
const commands: Command[] = [];
const files = await recursiveReaddir(
    path.join(dirname(import.meta.url), "..", "commands"),
);
for (const file of files) {
    if (!file.endsWith(".js") || file.endsWith(".template.js")) continue;
    logger.debug(`Loading command ${file}`);
    commands.push(
        (
            await import(
                path.join(dirname(import.meta.url), "..", "commands", file)
            )
        ).default as Command,
    );
}

export const processMessageCommand = async (message: Message) => {
    logger.debug(`Received message: ${message.content?.substring(0, 100)}`);

    if (message.server && !(await enabledInServer(message.server, client))) {
        return logger.debug(
            `Refusing to interact with server ${message.server.id}`,
        );
    }

    if (
        message.webhook ||
        message.systemMessage ||
        !message.content ||
        message.author?.bot
    )
        return logger.debug("Ignoring message");

    if (!message.channel)
        return logger.info(
            "Ignoring message; We don't have the channel in cache",
        );

    const serverConfig = message.server
        ? await client.db.servers.findOne({ _id: message.server.id })
        : null;

    const prefix = serverConfig?.prefix || DEFAULT_PREFIX;
    let usedPrefix: string;

    let matchMessage = message.content;
    const match = matchMessage.match(RE_PREFIX_MENTION);
    if (match?.[0]) {
        matchMessage = matchMessage.replace(match[0], "");
        usedPrefix = match[0];
    } else if (matchMessage.startsWith(prefix)) {
        matchMessage = matchMessage.substring(prefix.length);
        usedPrefix = prefix;
    } else return logger.debug("Message doesn't start with prefix");

    const args = matchMessage.replace(/\s+/, " ").split(" ");
    const cmdName = args.shift();
    if (!cmdName) return logger.debug("Message doesn't specify a command name");

    const command = commands.find(
        (c) => c.name == cmdName?.toLowerCase() || c.aliases?.includes(cmdName),
    );
    if (!command) return logger.debug("Command doesn't exist");

    if (command.serverOnly && !message.server?.id) {
        return await message.reply(
            embedMsg(
                await t("server_only.body", message.author),
                await t("server_only.title", message.author),
                "ERROR",
            ),
        );
    }

    try {
        if (
            !(await userHasCommandPermissions(
                command,
                message.author!,
                message.member,
                message,
            ))
        ) {
            return logger.debug("User has no permission to use this command");
        }

        const context: CommandContext = {
            usedPrefix,
            serverConfig,
            serverPrefix: prefix,
            commandName: cmdName,
        };
        await command.run(message, args, client, context);
    } catch (e) {
        console.error(e);
        await message.channel?.sendMessage(
            embedMsg(
                `\`\`\`js\n${e}\n\`\`\`\n` +
                    `[${await t(
                        "command_failed.bug_report",
                        message.author,
                    )}](${META.newIssueUrl})\n` +
                    `[${await t(
                        "command_failed.support_server",
                        message.author,
                    )}](${META.supportServer})`,
                await t("command_failed.title", message.author),
                "ERROR",
            ),
        );
    }
};

const userHasCommandPermissions = async (
    command: Command,
    user: User,
    member?: ServerMember,
    message?: Message,
): Promise<boolean> => {
    if (!command.permissions) return true;

    if (
        command.permissions.includes("BotAdmin") &&
        (!user.id || !BOT_ADMINS.includes(user.id))
    ) {
        logger.info("User is not bot admin!");
        if (message?.channel?.havePermission("React")) {
            await message.react(
                IS_SELF_HOSTED
                    ? encodeURIComponent("❌")
                    : "01GYQSY2C5TNK8WCPT0XDC5V22",
            );
        }
        return false;
    }

    for (const perm of command.permissions.filter(
        (perm) => !Object.keys(BotPermission).includes(perm),
    )) {
        try {
            if (
                member &&
                !member.hasPermission(
                    member.server!,
                    perm as keyof typeof Permission,
                )
            ) {
                await message?.reply(
                    embedMsg(
                        await t("no_permission.body", message.author, {
                            permissions: command.permissions
                                .map((perm) => `\`${perm}\``)
                                .join(", "),
                        }),
                        await t("no_permission.title", message.author),
                        "ERROR",
                    ),
                );
                return false;
            }
        } catch (e) {
            throw new Error("Permission check failed\n=> " + e);
        }
    }
    return true;
};

client.on("messageCreate", processMessageCommand);

export { commands, userHasCommandPermissions };
