import i18next from "i18next";
import {
    EmbedMessageType,
    TranslationGender,
    dirname,
    embedMsg,
    sanitizeMessageContent,
    DBServer,
} from "lib";
import path from "path";
import resourcesToBackend from "i18next-resources-to-backend";
import { client, logger } from "../index.js";
import { Server, User } from "revolt.js";
import fs from "fs/promises";
import { META } from "../meta.js";

type LangFile = {
    [key: string]: {
        name: string;
        nameLocalised: string;
        emoji: string;
    };
};

const dir = path.join(
    dirname(import.meta.url),
    "../../..",
    "external",
    "translations",
);

const { langs }: { langs: LangFile } = JSON.parse(
    await fs.readFile(path.join(dir, "langs.json"), "utf-8"),
);

i18next.use(
    resourcesToBackend((lang: string, ns: string) => {
        logger.info(`Loading translation ${ns}/${lang}`);
        return import(path.join(dir, ns, `${lang}.json`), {
            assert: { type: "json" },
        }).catch((e) =>
            logger.error(`Failed to load translation ${lang}: ${e}`),
        );
    }),
);
i18next.init({ ns: ["bot"], supportedLngs: Object.keys(langs), lng: "en" });

await i18next.loadLanguages(Object.keys(langs));

export type TranslationData = { [key: string]: unknown } & {
    lang?: string;
    gender?: TranslationGender;
    count?: number;
};

export const t = async (
    key: string,
    user: User | undefined,
    data: TranslationData = {},
    // eslint-disable-next-line require-await
) => {
    const userConfig = user
        ? await client.db.users.findOne({ _id: user.id })
        : null;

    const language = data?.lang ?? userConfig?.language ?? "en";
    const gender =
        data?.gender ?? userConfig?.pronouns ?? TranslationGender.Neutral; // TODO infer this from user pronouns (once available)
    const genderSuffix =
        gender == TranslationGender.Female
            ? "_f"
            : gender == TranslationGender.Male
            ? "_m"
            : "";

    data["interpolation"] = { escapeValue: false };
    data["meta"] = {
        username: user?.username
            ? sanitizeMessageContent(user.username).trim()
            : "-",
        ...META,
    };

    // Try [provided language gendered] => [provided language neutral] => [english gendered] => [english neutral]
    let translation: string | null = null;
    if (gender != TranslationGender.Neutral) {
        translation = i18next.t(`${key}${genderSuffix}`, {
            lng: language,
            defaultValue: null,
            ...data,
        });
    }

    translation ??= i18next.t(key, {
        lng: language,
        defaultValue: null,
        ...data,
    });
    translation ??= i18next.t(`${key}${genderSuffix}`, {
        defaultValue: null,
        ...data,
    });
    translation ??= i18next.t(key, { defaultValue: null, ...data });

    return translation ?? `{{${key}}}`;
};

// t() but uses the language config of a server
export const tServer = async (
    key: string,
    server: Server | DBServer | undefined,
    user: User | undefined,
    data: { [key: string]: unknown } & {
        gender?: TranslationGender;
        count?: number;
    } = {},
) => {
    const language: string | undefined =
        server instanceof Server
            ? (await client.db.servers.findOne({ _id: server.id }))?.language
            : server?.language;

    return await t(key, user, { ...data, lang: language || "en" });
};

export const tEmbedMsg = async (
    user: User | undefined,
    bodyKey: string,
    titleKey: string,
    type?: EmbedMessageType,
    data: TranslationData = {},
) =>
    embedMsg(await t(bodyKey, user, data), await t(titleKey, user, data), type);

export const joinTranslatedDetails = async (
    keyPrefix: string,
    user: User | undefined,
    server: Server | DBServer | undefined,
    data: ({
        [key: string]: unknown;
    } & { key: string; item: string })[],
    separator = "\n",
): Promise<string> => {
    const result: string[] = [];
    for (const item of data) {
        if (item[item.item]) {
            if (server) {
                result.push(
                    await tServer(
                        `${keyPrefix}.${item.key}`,
                        server,
                        user,
                        item,
                    ),
                );
            } else {
                result.push(await t(`${keyPrefix}.${item.key}`, user, item));
            }
        }
    }
    return result.join(separator);
};

export { langs };
