import {
    DBServer,
    LoggedAttachment,
    LoggedMessage,
    MessageLogType,
    ServerLogCategory,
    enabledInServer,
    fetchUser,
} from "lib";
import { Message, TextEmbed, File, Channel, User } from "revolt.js";
import { dispatchLogMessage } from "./server_logs.js";
import { ulid } from "ulid";
import { client, logger } from "../index.js";
import { HydratedMessage } from "revolt.js/src/hydration";

export async function processMessageUpdate(
    message: Message,
    oldContent: string | undefined,
    config: DBServer,
) {
    await dispatchLogMessage(
        {
            _id: ulid(),
            category: ServerLogCategory.Message,
            type: MessageLogType.Edit,
            server: message.server?.id || "",
            channel: message.channelId,
            messageId: message.id,
            user: message.authorId || "",
            content: message.content!,
            oldContent,
        },
        client,
        config,
    );
}

export async function processMessageDelete(
    message: HydratedMessage,
    channel: Channel,
    author: User,
    config: DBServer,
) {
    await dispatchLogMessage(
        {
            _id: ulid(),
            category: ServerLogCategory.Message,
            type: MessageLogType.Delete,
            server: channel.serverId,
            channel: message.channelId,
            messageId: message.id,
            user: message.authorId!,
            files: message.attachments?.map(fileToLogAttachment) || [],
            content: message.content,
            author: {
                username: author.username,
                discriminator: author.discriminator,
                displayname: author.displayName,
                id: message.authorId,
                bot: author.bot?.owner,
            },
            embeds: message.embeds?.filter(
                (e) => e.type == "Text",
            ) as TextEmbed[],
        },
        client,
        config,
    );
}

export async function processMessageBulkDelete(
    messages: HydratedMessage[],
    channel: Channel,
    config?: DBServer,
) {
    await dispatchLogMessage(
        {
            _id: ulid(),
            category: ServerLogCategory.Message,
            type: MessageLogType.DeleteBulk,
            server: channel.serverId || "",
            channel: channel.id,
            messages: await Promise.all(
                messages.map(async (message) => {
                    const author = message.authorId
                        ? await fetchUser(message.authorId, client)
                        : null;
                    return {
                        author: {
                            id: message.authorId,
                            username: author?.username,
                            discriminator: author?.discriminator,
                            displayname: author?.displayName,
                            bot: author?.bot?.owner,
                        },
                        content: message.content,
                        files:
                            message.attachments?.map(fileToLogAttachment) || [],
                        embeds: message.embeds?.filter(
                            (embed) => embed.type == "Text",
                        ) as TextEmbed[],
                    } as LoggedMessage;
                }),
            ),
        },
        client,
        config,
    );
}

const fileToLogAttachment = (file: File): LoggedAttachment => ({
    id: file.id,
    autumnTag: file.tag,
    size: file.size,
    spoiler: file.isSpoiler,
    contentType: file.contentType,
    filename: file.filename,
    type: file.metadata.type,
});

client.on("messageUpdate", async (message, oldMessage) => {
    logger.debug(`Received edit event for ${message.id}`);
    try {
        if (!message.server?.id) return logger.debug("Message has no server");
        if (!(await enabledInServer(message.server, client)))
            return logger.debug("Event happened in ignored server");
        if (message.content == oldMessage.content) {
            return logger.debug("Message content hasn't changed");
        }

        const config = await client.db.servers.findOne({
            _id: message.server.id,
        });

        if (config?.logs?.messages) {
            await processMessageUpdate(message, oldMessage.content, config);
        }
    } catch (e) {
        logger.error(`Error processing message edit ${message.id}`);
        console.log(e);
    }
});

client.on("messageDelete", async (message) => {
    logger.debug(`Received delete event for ${message.id}`);
    try {
        if (!message.channelId || !message.authorId)
            return logger.debug("No channel id or author id");
        const channel = client.channels.get(message.channelId);
        if (!channel?.serverId) return logger.debug("No channel or server id");
        if (!(await enabledInServer({ id: channel.serverId }, client)))
            return logger.debug("Event happened in ignored server");
        const author = await fetchUser(message.authorId, client);
        if (!author) return logger.debug("No author");

        if (message.systemMessage) {
            return logger.debug(
                `${message.id} is a system message, not logging deletion`,
            );
        }

        const config = await client.db.servers.findOne({
            _id: channel.serverId,
        });

        if (config?.logs?.messages) {
            await processMessageDelete(message, channel, author, config);
        }
    } catch (e) {
        logger.error(`Error processing delete event for ${message.id}`);
        console.log(e);
    }
});

client.on("messageDeleteBulk", async (messages, channel) => {
    logger.debug(
        `Received bulk delete event in ${channel?.id || "an unknown channel"}`,
    );
    try {
        if (!channel?.serverId) return logger.debug("No channel or server ID");
        if (!(await enabledInServer({ id: channel.serverId }, client)))
            return logger.debug("Event happened in ignored server");
        const config = await client.db.servers.findOne({
            _id: channel.serverId!,
        });

        if (config?.logs?.messages) {
            await processMessageBulkDelete(messages, channel, config);
        }
    } catch (e) {
        logger.error(
            `Error processing bulk delete in ${
                channel?.id || "an unknown channel"
            }`,
        );
        console.log(e);
    }
});
