import {
    CreateInfractionReq,
    CreateInfractionRes,
    DBInfraction,
    GetMemberReq,
    GetMemberRes,
    GetMutualServersReq,
    GetMutualServersRes,
    GetServerDetailsReq,
    GetServerDetailsRes,
    InfractionType,
    ModActionType,
    ServerLogCategory,
    canModifyServerSettings,
    connectRedis,
    createInfraction,
    fetchMember,
    fetchUser,
    generateInfractionId,
    memberIsModerator,
    redisIPCHandler,
} from "lib";
import { Permission } from "revolt.js/lib/cjs/permissions/definitions.js";
import { client, environment } from "../index.js";
import { dispatchLogMessage } from "./server_logs.js";
import { ulid } from "ulid";

const rcvRedis = await connectRedis(environment.REDIS_URI);

redisIPCHandler<GetMemberReq, GetMemberRes>(
    "getMemberInfo",
    client.redis,
    rcvRedis,
    async (req) => {
        const member = await fetchMember(req.user, req.server, client);
        if (!member) return { id: req.user, isMember: false };

        const perms = member.getPermissions(member.server!);

        return {
            id: member.id.user,
            isMember: true,
            isModerator: await memberIsModerator(member),
            isAdmin: canModifyServerSettings(member),
            permissions: Object.entries(Permission)
                .filter((val) => val[1] & perms)
                .map((val) => val[0]) as (keyof typeof Permission)[],
            ranking: member.ranking,
        };
    },
);

redisIPCHandler<CreateInfractionReq, CreateInfractionRes>(
    "createInfraction",
    client.redis,
    rcvRedis,
    async (req) => {
        const server = client.servers.get(req.server);
        if (!server) throw new Error("Unknown server");

        // const target = await fetchMember(req.target, req.server, client);
        const mod = await fetchMember(req.mod, req.server, client);
        if (!mod) throw new Error("Infraction author is not a server member");
        if (!mod.hasPermission(mod.server!, "KickMembers"))
            throw "Unauthorised";
        // if (target && mod.ranking >= target.ranking)

        const infraction: DBInfraction = {
            _id: {
                infraction: await generateInfractionId(server, client.db),
                server: server.id,
            },
            user: req.target,
            context: null,
            date: new Date(Date.now()),
            moderator: req.mod,
            reason: req.reason,
            type: InfractionType.Warn,
        };

        const count = await createInfraction(infraction, server, client);
        await dispatchLogMessage(
            {
                _id: ulid(),
                category: ServerLogCategory.ModAction,
                type: ModActionType.CreateInfraction,
                infraction: infraction._id,
                server: server.id,
            },
            client,
        );
        return { count };
    },
);

redisIPCHandler<GetMutualServersReq, GetMutualServersRes>(
    "getMutualServers",
    client.redis,
    rcvRedis,
    async (msg) => {
        const user = await fetchUser(msg.user, client);
        if (!user) throw "Cannot find user";
        const mutual = await user.fetchMutual();

        return await Promise.all(
            mutual.servers
                .map((server) => client.servers.get(server))
                .filter((server) => server)
                .map(async (server) => ({
                    id: server!.id,
                    name: server!.name,
                    icon: server?.animatedIconURL,
                    discoverable: !!server?.discoverable,
                    isModerator: await memberIsModerator(
                        await fetchMember(user.id, server!.id, client),
                    ),
                })),
        );
    },
);

redisIPCHandler<GetServerDetailsReq, GetServerDetailsRes>(
    "getServerDetails",
    client.redis,
    rcvRedis,
    // eslint-disable-next-line require-await
    async (msg) => {
        const server = client.servers.get(msg.server);
        if (!server) throw "Server not found";
        const member = await fetchMember(msg.user, msg.server, client);
        if (!member) throw "Not a member of this server";

        return {
            id: server.id,
            name: server.name,
            icon: server.animatedIconURL,
            discoverable: !!server?.discoverable,
            isModerator: await memberIsModerator(member),
        };
    },
);
