/* eslint-disable @typescript-eslint/no-explicit-any */
import {
    AutomodClient,
    DBServerLog,
    InfractionType,
    MessageLogType,
    ModActionType,
    ServerLogCategory,
    countUserInfractions,
    embed,
    fetchUser,
    hideNewlines,
    limitLength,
    sanitizeMessageContent,
    DBServer,
} from "lib";
import Chroma from "chroma-js";
import { client, logger } from "../index.js";
import { API } from "revolt.js";
import { Permission } from "revolt.js/lib/esm/permissions/definitions.js";
import { joinTranslatedDetails, tServer } from "./i18n.js";

const DELETED_CONTENT_MAX_LENGTH = 800;
const EDITED_CONTENT_MAX_LENGTH = 500;

export async function dispatchLogMessage(
    message: DBServerLog,
    client: AutomodClient,
    config?: DBServer | null,
) {
    logger.info(`Dispatching log message ${message._id}`);
    try {
        if (!config) {
            config =
                (await client.db.servers.findOne({ _id: message.server })) ||
                undefined;
        }

        let channelId: string | undefined = undefined;
        let delivery: DBServerLog["delivery"] = undefined;

        switch (message.category) {
            case ServerLogCategory.ModAction:
                channelId = config?.logs?.moderation;
                break;
            case ServerLogCategory.Message:
                channelId = config?.logs?.messages;
                break;
            default:
                break;
        }

        if (channelId) {
            try {
                const channel =
                    client.channels.get(channelId!) ||
                    (await client.channels.fetch(channelId!));
                for (const perm of [
                    "SendMessage",
                    "SendEmbeds",
                    "UploadFiles",
                ] as (keyof typeof Permission)[]) {
                    if (!channel.havePermission(perm))
                        throw `Missing permission ${perm} for channel ${channelId}`;
                }

                const embed = await createLogMessage(message, config, client);
                const colour = await getLogMessageColour(channel.id);
                await channel.sendMessage({ embeds: [{ ...embed, colour }] });
                // TODO batch messages if multiple logs come through at the same time
            } catch (e) {
                logger.error(`Delivery for log message ${message._id} failed`);
                console.log(e);
                delivery = { state: "failed", error: `${e}` };
            }
        } else delivery = { state: "no_channel" };

        delivery ??= { state: "success" };
        await client.db.serverLogs.insertOne({ ...message, delivery });
    } catch (e) {
        logger.error("Error while dispatching server log!");
        console.log(e);
    }
}

async function createLogMessage(
    log: DBServerLog,
    config: DBServer | undefined,
    client: AutomodClient,
): Promise<API.SendableEmbed> {
    const eventWebLink = "https://todo";

    switch (log.category) {
        case ServerLogCategory.ModAction: {
            const infraction = await client.db.infractions.findOne({
                _id: log.infraction,
            });
            if (!infraction) {
                throw new Error(`Infraction ${log.infraction} not found`);
            }

            const user = await fetchUser(infraction.user, client);
            const mod = await fetchUser(infraction.moderator, client);
            const actingUser = log.user
                ? await fetchUser(log.user, client)
                : null;
            const warnCount = await countUserInfractions(
                infraction.user,
                log.server,
                client,
            );
            const title = await tServer(
                `server_logs.${log.type}.title`,
                config,
                user || undefined,
            );

            let description: string;
            switch (log.type) {
                case ModActionType.CreateInfraction: {
                    description =
                        "### " +
                        (await tServer(
                            `server_logs.${log.type}.message.${infraction.type}`,
                            config,
                            user || undefined,
                            {
                                moderator: `\`${mod?.username || "Unknown"}\``,
                                target: `\`${user?.username || "Unknown"}\``,
                            },
                        )) +
                        "\n" +
                        (await joinTranslatedDetails(
                            "server_logs.details",
                            user || undefined,
                            config,
                            [
                                {
                                    key: "infraction_id",
                                    item: "id",
                                    id: infraction._id.infraction,
                                },
                                {
                                    key: "target_user",
                                    item: "id",
                                    id: infraction.user,
                                    username: user
                                        ? `${user.username}#${user.discriminator}`
                                        : "*Unknown*",
                                },
                                {
                                    key: "moderator",
                                    item: "id",
                                    id: infraction.moderator,
                                    username: mod
                                        ? `${mod.username}#${mod.discriminator}`
                                        : "*Unknown*",
                                },
                                {
                                    key: "ban_expiry",
                                    item: "ts",
                                    ts:
                                        infraction.type ==
                                        InfractionType.BanTemporary
                                            ? `<t:${Math.round(
                                                  (
                                                      infraction as any
                                                  ).expires.getTime() / 1000,
                                              )}:d>`
                                            : null,
                                },
                                {
                                    key: "reason",
                                    item: "reason",
                                    reason: hideNewlines(
                                        infraction.reason ||
                                            (await tServer(
                                                "infractions.unknown_reason",
                                                config,
                                                user || undefined,
                                            )),
                                    ),
                                },
                                {
                                    key: "warn_number",
                                    item: "count",
                                    count: warnCount,
                                },
                            ],
                        ));
                    break;
                }
                case ModActionType.RemoveInfraction:
                case ModActionType.EditInfraction:
                case ModActionType.TempBanExpire: {
                    description =
                        "#### " +
                        (await tServer(
                            `server_logs.${log.type}.content`,
                            config,
                            user || undefined,
                            {
                                username: user
                                    ? user.displayName
                                    : infraction.user,
                            },
                        )) +
                        "\n" +
                        (await joinTranslatedDetails(
                            "server_logs.details",
                            user || undefined,
                            config,
                            log.type == ModActionType.RemoveInfraction
                                ? [
                                      {
                                          key: "infraction_id",
                                          item: "id",
                                          id: infraction._id.infraction,
                                      },
                                      {
                                          key: "target_user",
                                          item: "id",
                                          id: infraction.user,
                                          username: user
                                              ? `${user.username}#${user.discriminator}`
                                              : "*Unknown*",
                                      },
                                      {
                                          key: "moderator",
                                          item: "id",
                                          id: log.user,
                                          username: actingUser
                                              ? `${actingUser.username}#${actingUser.discriminator}`
                                              : "*Unknown*",
                                      },
                                  ]
                                : log.type == ModActionType.EditInfraction
                                ? [
                                      {
                                          key: "infraction_id",
                                          item: "id",
                                          id: infraction._id.infraction,
                                      },
                                      {
                                          key: "target_user",
                                          item: "id",
                                          id: infraction.user,
                                          username: user
                                              ? `${user.username}#${user.discriminator}`
                                              : "*Unknown*",
                                      },
                                      {
                                          key: "moderator",
                                          item: "id",
                                          id: log.user,
                                          username: actingUser
                                              ? `${actingUser.username}#${actingUser.discriminator}`
                                              : "*Unknown*",
                                      },
                                      {
                                          key: "old_reason",
                                          item: "reason",
                                          reason: (log as any).oldReason,
                                      },
                                      {
                                          key: "new_reason",
                                          item: "reason",
                                          reason: infraction.reason,
                                      },
                                  ]
                                : [
                                      {
                                          key: "infraction_id",
                                          item: "id",
                                          id: infraction._id.infraction,
                                      },
                                      {
                                          key: "target_user",
                                          item: "id",
                                          id: infraction.user,
                                          username: user
                                              ? `${user.username}#${user.discriminator}`
                                              : "*Unknown*",
                                      },
                                  ],
                        ));
                    break;
                }
                default:
                    return embed(
                        JSON.stringify(log, null, 4),
                        "Unknown log action type",
                    );
            }

            return embed(description, title);
        }
        case ServerLogCategory.Message: {
            let description: string;
            let footer = "";
            let header = "";

            if (log.type != MessageLogType.DeleteBulk) {
                header = await tServer(
                    "server_logs.message.info.header",
                    config,
                    undefined,
                    {
                        channel: `<#${log.channel}>`,
                        user: log.user ? `<@${log.user}>` : "Unknown",
                        url: `/server/${log.server}/channel/${log.channel}/${log.messageId}`,
                    },
                );
                footer =
                    "##### " +
                    (await tServer(
                        "server_logs.message.info.embeds",
                        config,
                        undefined,
                        { count: (log as any).embeds?.length || 0 },
                    )) +
                    " &bull; " +
                    (await tServer(
                        "server_logs.message.info.attachments",
                        config,
                        undefined,
                        { count: (log as any).files?.length || 0 },
                    )) +
                    " &bull; " +
                    `[${await tServer(
                        "server_logs.message.info.show_more",
                        config,
                        undefined,
                    )}](${eventWebLink})`;
            }

            switch (log.type) {
                case MessageLogType.Edit: {
                    description = header;
                    description += "\n#### ";
                    description += await tServer(
                        "server_logs.message.edit.old_content",
                        config,
                        undefined,
                    );
                    if (!log.oldContent) {
                        description +=
                            "\n" +
                            (await tServer(
                                "server_logs.message.info.content_unknown",
                                config,
                                undefined,
                            ));
                    } else {
                        description +=
                            "\n>" +
                            sanitizeMessageContent(
                                limitLength(
                                    log.oldContent,
                                    EDITED_CONTENT_MAX_LENGTH,
                                ),
                            )
                                .trim()
                                .replaceAll("\n", "\n>");
                    }

                    description += "\n\n#### ";
                    description += await tServer(
                        "server_logs.message.edit.new_content",
                        config,
                        undefined,
                    );
                    description +=
                        "\n>" +
                        sanitizeMessageContent(
                            limitLength(
                                log.content || "",
                                EDITED_CONTENT_MAX_LENGTH,
                            ),
                        )
                            .trim()
                            .replaceAll("\n", "\n>");

                    description += "\n\n" + footer;
                    break;
                }
                case MessageLogType.Delete: {
                    description = header;
                    description += "\n#### ";
                    description += await tServer(
                        "server_logs.message.delete.content",
                        config,
                        undefined,
                    );
                    if (!log.content) {
                        description +=
                            "\n" +
                            (await tServer(
                                "server_logs.message.info.content_unknown",
                                config,
                                undefined,
                            ));
                    } else {
                        description +=
                            "\n>" +
                            sanitizeMessageContent(
                                limitLength(
                                    log.content,
                                    DELETED_CONTENT_MAX_LENGTH,
                                ),
                            )
                                .trim()
                                .replaceAll("\n", "\n>");
                    }

                    description += "\n\n" + footer;
                    break;
                }
                case MessageLogType.DeleteBulk: {
                    description = await tServer(
                        "server_logs.message.bulk_delete.content",
                        config,
                        undefined,
                        {
                            count: log.messages.length,
                            channel: `<#${log.channel}>`,
                            url: eventWebLink,
                        },
                    );
                    break;
                }
                default: {
                    throw new Error(
                        `Unknown MessageLogType: ${(log as any).type}`,
                    );
                }
            }

            return embed(
                description,
                await tServer(
                    `server_logs.message.${log.type}.title`,
                    config,
                    undefined,
                ),
            );
        }
        default:
            return embed(
                JSON.stringify(log, null, 4),
                "You should never see this in production",
            );
    }
}

async function getLogMessageColour(channel: string) {
    const rotation = await client.redis.INCRBY(`log_gradient.${channel}`, 15);
    return Chroma.hsv(rotation % 360, 0.5, 0.7).hex("rgb");
}
