import {
    AutomodClient,
    DBInfraction,
    ModActionType,
    ServerLogCategory,
    embed,
    embedMsg,
    fetchMember,
    yesNoMessage,
    DBServer,
} from "lib";
import { API, Message, ServerMember, User } from "revolt.js";
import { t } from "../modules/i18n.js";
import { ulid } from "ulid";
import { client } from "../index.js";
import { dispatchLogMessage } from "../modules/server_logs.js";

const MAX_USERS = 5;

export async function checkConditions(
    message: Message,
    client: AutomodClient,
    users: (User | { id: string })[],
    notFound: number,
    certain: boolean,
): Promise<boolean> {
    if (!message.channel) throw new Error("channel is not set");
    if (!message.authorId) throw new Error("authorId is not set");
    if (!users.length) {
        await message.reply(
            embedMsg(
                await t("infractions.no_user", message.author, {
                    count: notFound,
                }),
                await t("infractions.title_create", message.author),
                "ERROR",
            ),
        );
        return false;
    }

    if (notFound > 0) {
        const result = await yesNoMessage(
            message.channel,
            client,
            message.authorId,
            await t("target_uncertain.not_all_found", message.author, {
                count: users.length,
                users: users
                    .map((u) =>
                        u instanceof User
                            ? `\`${u.username}#${u.discriminator}\``
                            : `\`${u.id}\``,
                    )
                    .join(", "),
            }),
            await t("target_uncertain.title", message.author),
            await t("target_uncertain.confirmed", message.author),
            await t("target_uncertain.canceled", message.author),
        );

        return result;
    } else if (!certain) {
        const result = await yesNoMessage(
            message.channel,
            client,
            message.authorId,
            await t("target_uncertain.body", message.author, {
                count: users.length,
                users: users
                    .map((u) =>
                        u instanceof User
                            ? `\`${u.username}#${u.discriminator}\``
                            : `\`${u.id}\``,
                    )
                    .join(", "),
            }),
            await t("target_uncertain.title", message.author),
            await t("target_uncertain.confirmed", message.author),
            await t("target_uncertain.canceled", message.author),
        );

        return result;
    }

    if (users.length > MAX_USERS) {
        await message.reply(
            embedMsg(
                await t("infractions.too_many_users", message.author),
                await t("infractions.title_create", message.author),
                "ERROR",
            ),
        );
        return false;
    }

    return true;
}

/**
 * Checks whether a specific user can be removed with /kick or /ban, and returns the error as message embed if not.
 */
export async function checkRemovalEligibility(
    message: Message,
    target: ServerMember,
    targetUser: User,
    client: AutomodClient,
    isBan: boolean,
): Promise<API.SendableEmbed | void> {
    if (targetUser.id == client.user?.id) {
        return embed(
            await t("infractions.death_threats", message.author),
            await t("infractions.title_create", message.author),
            "WARN",
        );
    }

    if (!target.inferiorTo(message.member!)) {
        return embed(
            await t("permissions.target_higher_than_you", message.author),
            await t("infractions.title_create", message.author),
            "WARN",
        );
    }

    if (
        !target.inferiorTo(message.server!.member!) ||
        !message.server!.havePermission(isBan ? "BanMembers" : "KickMembers")
    ) {
        return embed(
            await t(
                `permissions.bot_cant_${isBan ? "ban" : "kick"}`,
                message.author,
            ),
            await t("infractions.title_create", message.author),
            "WARN",
        );
    }
}

export const sendInfractionLog = async (
    infraction: DBInfraction,
    config: DBServer | null,
) =>
    await dispatchLogMessage(
        {
            _id: ulid(),
            category: ServerLogCategory.ModAction,
            type: ModActionType.CreateInfraction,
            server: infraction._id.server,
            infraction: infraction._id,
            user: infraction.user,
        },
        client,
        config,
    );

export async function timeoutMember(
    user: string,
    server: string,
    duration: Date,
) {
    await client.db.members.updateOne(
        { _id: { user, server } },
        { $set: { timeout: duration } },
        { upsert: true },
    );

    const member = await fetchMember(user, server, client);
    if (member && member.server?.havePermission("TimeoutMembers")) {
        await member.edit({ timeout: duration.toISOString() });
    }
}

export async function clearMemberTimeout(user: string, server: string) {
    await client.db.members.updateOne(
        { _id: { user, server } },
        { $unset: { timeout: true } },
        { upsert: true },
    );

    const member = await fetchMember(user, server, client);
    if (member && member.server?.havePermission("TimeoutMembers")) {
        await member.edit({ remove: ["Timeout"] });
    }
}
