import EventEmitter from "events";
import { ulid } from "ulid";
import { createLogger } from "./logger";
import { AutomodRedis } from "./db";

type RedisIPCMessage<T> = { id: string } & (
    | { message: T; error?: false }
    | { message: any; error: true }
);

/**
 * Fancy wrapper for redis pubsub that lets us do request/response communication.
 * We use two different channels, `<channel>_req` and `<channel>_res`. Incoming messages
 * are received on `req` and the response is sent to `res`.
 */
export function redisIPCHandler<REQ, RES>(
    channel: string,
    sndRedis: AutomodRedis,
    rcvRedis: AutomodRedis,
    cb: (message: REQ) => Promise<RES>,
) {
    const logger = createLogger();
    rcvRedis.SUBSCRIBE(`${channel}_req`, async (msg) => {
        try {
            const json: RedisIPCMessage<REQ> = JSON.parse(msg);

            let message: RES;

            try {
                message = await cb(json.message);
            } catch (e) {
                await sndRedis.PUBLISH(
                    `${channel}_res`,
                    JSON.stringify({
                        id: json.id,
                        message: "" + e,
                        error: true,
                    } as RedisIPCMessage<RES>),
                );
                return;
            }

            const res: RedisIPCMessage<RES> = {
                id: json.id,
                message,
            };
            await sndRedis.PUBLISH(`${channel}_res`, JSON.stringify(res));
        } catch (e) {
            logger.error(`Error processing pubsub message ${channel}`);
            console.log(`${e}`);
        }
    });
}

export function createRedisIPCSender<REQ, RES>(
    channel: string,
    sndRedis: AutomodRedis,
    rcvRedis: AutomodRedis,
): (message: REQ) => Promise<RES> {
    const emitter = new EventEmitter();

    rcvRedis.SUBSCRIBE(`${channel}_res`, (msg) => {
        const data: RedisIPCMessage<RES> = JSON.parse(msg);
        emitter.emit(data.id, data);
    });

    return (message) => {
        const id = ulid();
        sndRedis.PUBLISH(
            `${channel}_req`,
            JSON.stringify({
                id,
                message,
            } as RedisIPCMessage<REQ>),
        );

        return new Promise<RES>((resolve, reject) => {
            const timeout = setTimeout(() => {
                emitter.removeAllListeners(id);
                reject(new Error("Request timed out"));
            }, 5000);
            emitter.on(id, (data: RedisIPCMessage<RES>) => {
                clearTimeout(timeout);

                if (data.error) {
                    reject(data.message);
                } else {
                    resolve(data.message);
                }
            });
        });
    };
}
