import { fileURLToPath } from "url";
import { API, Channel, Message, Server, ServerMember, User } from "revolt.js";
import path from "path";
import fs from "fs/promises";
import crypto from "crypto";
import { AutomodClient } from "../types/AutomodClient";
import { RE_MENTION_USER, RE_ULID } from "./regex.js";
import { DBInfraction } from "../types/database/infraction";
import { EventProtocol } from "revolt.js/src/events";
import { OrID } from "../types/OrID";
import { AutomodDatabase } from "./db";
import { createLogger } from "./logger";

/**
 * Replacement for __dirname which the nodejs people so graciously took away.
 * Pass `import.meta.url` as parameter
 * @param url
 * @returns
 */
export const dirname = (url: string) => path.dirname(fileURLToPath(url));

/**
 * Asynchronously returns all files in a given directory, including subdirectories.
 * Only returns files, not directories.
 */
export const recursiveReaddir = async (dir: string): Promise<string[]> => {
    const files = await fs.readdir(dir);
    const paths: string[] = [];
    for (const file of files) {
        const fullfile = path.join(dir, file);
        const stat = await fs.stat(fullfile);
        if (stat.isSymbolicLink()) continue;
        if (stat.isDirectory()) {
            paths.push(
                ...(await recursiveReaddir(fullfile)).map((f) =>
                    path.join(file, f),
                ),
            );
        } else {
            paths.push(file);
        }
    }

    return paths;
};

export type EmbedMessageType = "INFO" | "SUCCESS" | "WARN" | "ERROR";

export const embed = (
    content: string,
    title?: string,
    type?: EmbedMessageType,
): API.SendableEmbed => {
    const colors = {
        SUCCESS: "var(--success)",
        INFO: "var(--status-streaming)",
        WARN: "var(--warning)",
        ERROR: "var(--error)",
    };

    return {
        colour: colors[type || "INFO"],
        description: content,
        title: title,
    };
};

export const embedMsg = (
    content: string,
    title: string,
    type?: EmbedMessageType,
) => ({
    embeds: [embed(content, title, type)],
});

/**
 * Deduplicates the input array
 */
export function dedupe<T>(input: T[]): T[] {
    // This also makes sure we don't mutate the input array
    const result: T[] = [];
    for (const item of input) {
        if (!result.find((i) => Object.is(item, i))) result.push(item);
    }

    return result;
}

/**
 * Get or fetch a server member, return null if not found
 */
export async function fetchMember(
    user: string,
    server: string,
    client: AutomodClient,
): Promise<ServerMember | null> {
    try {
        return (
            client.serverMembers.getByKey({
                server,
                user,
            }) || (await client.serverMembers.fetch(server, user))
        );
    } catch (e) {
        return null;
    }
}

/**
 * Get or fetch a server member, return null if not found
 */
export async function fetchUser(
    user: string,
    client: AutomodClient,
): Promise<User | null> {
    try {
        return client.users.get(user) || (await client.users.fetch(user));
    } catch (e) {
        return null;
    }
}

/**
 * Resolve user input to an user object, e.g. <@mention>, user ID, or user name.
 * If a server ID is passed to `returnMember`, a member object will be returned instead.
 * If `certain` is false, we should ask for confirmation before running an action against the resolved user.
 */
export async function resolveUser<T extends string | undefined>(
    input: string,
    client: AutomodClient,
    serverId: T,
): Promise<{
    user: User | null;
    member: T extends string ? ServerMember | null : null;
    certain: boolean;
}> {
    if (
        new RegExp(RE_ULID).test(input) ||
        new RegExp(RE_MENTION_USER).test(input)
    ) {
        input = input.replace("<@", "").replace(">", "");
        return {
            user: client.users.get(input) || (await fetchUser(input, client)),
            member: (serverId
                ? await fetchMember(input, serverId, client)
                : null) as never,
            certain: true,
        };
    }

    // TODO use https://github.com/revoltchat/rfcs/pull/2 when merged
    let user = client.users.find(
        (u) => u.username.toLowerCase() == input.toLowerCase(),
    );
    if (user && serverId) {
        let member = client.serverMembers.getByKey({
            server: serverId,
            user: user.id,
        });
        if (!member) return { user, member: null, certain: false };
        return {
            user,
            member: member as never,
            certain: false,
        };
    }

    return { user: null, member: null, certain: true };
}

/**
 * Takes an input message and splits off the mentioned users.
 * Example: /warn <@a>,<@b> blabla -> users: [a, b], reason: "blablabla"
 * If the message is in reply to one or more other messages, the authors of those messages
 * will be used: /warn <@a>,<@b> blabla -> users: (from replies), reason: "<@a>,<@b> blabla"
 */
export async function separateUser<T extends boolean>(
    message: Message,
    args: string[],
    client: AutomodClient,
    returnUnknownIds: T,
): Promise<{
    users: (T extends true ? User | { id: string } : User)[];
    members?: ServerMember[];
    reason: string | null;
    certain: boolean;
    notFound: number;
}> {
    const users: (T extends true ? User | { id: string } : User)[] = [];
    const members: ServerMember[] = [];
    let notFound: number = 0;

    if (message.replyIds?.length) {
        // This will not work if replying to a message in a different channel (which the API allows, for some reason).
        const repliedMessages = (
            await Promise.all(
                message.replyIds.map(
                    async (id) =>
                        client.messages.get(id) ||
                        (await client.messages.fetch(message.channelId, id)),
                ),
            )
        ).filter((m) => m.authorId && !m.webhook?.id);

        for (const m of repliedMessages) {
            if (m.author) {
                users.push(m.author as never);
                if (m.member) {
                    members.push(m.member);
                } else if (m.channel?.serverId) {
                    const res = await resolveUser(
                        m.authorId!,
                        client,
                        message.server?.id,
                    );
                    if (res.member) members.push(res.member);
                }
            } else {
                const res = await resolveUser(
                    m.authorId!,
                    client,
                    message.server?.id,
                );
                if (res.user) users.push(res.user as never);
                else {
                    if (returnUnknownIds && m.authorId)
                        users.push({ id: m.authorId } as never);
                    else notFound++;
                }
                if (res.member) members.push(res.member);
            }
        }

        return {
            users: dedupe(users),
            members: dedupe(members),
            reason: args.join(" ") || null,
            certain: false,
            notFound,
        };
    }

    if (!args.length)
        return { users: [], reason: null, certain: true, notFound };

    let certain = true;
    const userInput = args[0].split(",");
    for (const input of userInput) {
        const res = await resolveUser(input, client, message.server?.id);
        if (!res.certain) certain = false;
        if (res.user) {
            users.push(res.user as never);
            if (res.member) {
                members.push(res.member);
            }
        } else {
            if (returnUnknownIds) {
                let id: string | null = null;
                if (new RegExp(RE_ULID).test(input)) id = input;
                else if (new RegExp(RE_MENTION_USER).test(input))
                    id = input.replace("<@", "").replace(">", "");
                else {
                    notFound++;
                    continue;
                }

                users.push({ id } as never);
            } else {
                notFound++;
            }
        }
    }

    return {
        users: dedupe(users),
        members: members.length ? dedupe(members) : undefined,
        reason: args.slice(1).join(" ") || null,
        certain,
        notFound,
    };
}

/**
 * Attempts to escape a message's markdown content (qoutes, headers, **bold** / *italic*, etc)
 */
export function sanitizeMessageContent(msg: string): string {
    let str = "";
    for (let line of msg.split("\n")) {
        line = line.trim();

        if (
            line.startsWith("#") || // headers
            line.startsWith(">") || // quotes
            line.startsWith("|") || // tables
            line.startsWith("*") || // unordered lists
            line.startsWith("-") || // ^
            line.startsWith("+") // ^
        ) {
            line = `\\${line}`;
        }

        // Ordered lists can't be escaped using `\`,
        // so we just put an invisible character \u200b
        if (/^[0-9]+[)\.].*/gi.test(line)) {
            line = `\u200b${line}`;
        }

        for (const char of ["_", "!!", "~", "`", "*", "^", "$"]) {
            line = line.replace(
                new RegExp(`(?<!\\\\)\\${char}`, "g"),
                `\\${char}`,
            );
        }

        // Mentions
        line = line.replace(/<@/g, `<\\@`);

        str += line + "\n";
    }

    return str;
}

export async function createInfraction(
    infraction: DBInfraction,
    server: Server,
    client: AutomodClient,
): Promise<number> {
    // TODO send message to log channel and message target user
    await client.db.infractions.insertOne(infraction);
    return await countUserInfractions(
        infraction.user,
        infraction._id.server,
        client,
    );
}

export const countUserInfractions = async (
    user: string,
    server: string,
    client: AutomodClient,
) =>
    client.db.infractions.countDocuments({
        "_id.server": server,
        user: user,
        revoked: { $ne: true },
    });

export const yesNoMessage = (
    channel: Channel,
    client: AutomodClient,
    allowedUser: string,
    message: string,
    title?: string,
    messageYes?: string,
    messageNo?: string,
): Promise<boolean> =>
    new Promise(async (resolve, reject) => {
        const EMOJI_YES = "✅",
            EMOJI_NO = "❌";
        try {
            const msg = await channel.sendMessage({
                embeds: [
                    {
                        colour: "var(--status-streaming)",
                        title: title,
                        description: message,
                    },
                ],
                interactions: {
                    reactions: [EMOJI_YES, EMOJI_NO],
                    restrict_reactions: true,
                },
            });

            let destroyed = false;
            const cb = async (packet: EventProtocol<1>["server"]) => {
                if (packet.type != "MessageReact") return;
                if (packet.id != msg.id) return;
                if (packet.user_id != allowedUser) return;

                switch (packet.emoji_id) {
                    case EMOJI_YES:
                        client.events.removeListener("event", cb);
                        destroyed = true;
                        resolve(true);
                        msg.edit({
                            embeds: [
                                {
                                    colour: "var(--success)",
                                    title: title,
                                    description: `${EMOJI_YES} ${
                                        messageYes ?? "Confirmed!"
                                    }`,
                                },
                            ],
                        }).catch((e) => console.error(e));
                        break;

                    case EMOJI_NO:
                        client.events.removeListener("event", cb);
                        destroyed = true;
                        resolve(false);
                        msg.edit({
                            embeds: [
                                {
                                    colour: "var(--error)",
                                    title: title,
                                    description: `${EMOJI_NO} ${
                                        messageNo ?? "Cancelled."
                                    }`,
                                },
                            ],
                        }).catch((e) => console.error(e));
                        break;
                }
            };
            client.events.on("event", cb);

            setTimeout(() => {
                if (!destroyed) {
                    resolve(false);
                    client.events.removeListener("event", cb);
                    msg.edit({
                        embeds: [
                            {
                                colour: "var(--error)",
                                title: title,
                                description: `${EMOJI_NO} Timed out`,
                            },
                        ],
                    }).catch((e) => console.error(e));
                }
            }, 30000);
        } catch (e) {
            reject(e);
        }
    });

export const getDmChannel = async (user: User, client: AutomodClient) => {
    return (
        Array.from(client.channels.values()).find(
            (c: Channel) =>
                c.type == "DirectMessage" &&
                c.recipient?.id == (user as User).id,
        ) || (await user.openDM())
    );
};

/**
 * Reads a command argument array, filters the first argument if it's a duration timestamp and returns its length in milliseconds.
 * The rest is returned as string.
 * Example: `["10m30s", "abcd", "amogus"]` -> `{ time: 630000, text: "abcd amogus" }`
 * Uses bigints internally so stupidly large user input can't overflow it.
 */
export function splitTimeInput(input: string[]): {
    time: number;
    text: string;
} {
    const args = [...input]; // Don't want to mutate the input array

    let duration = BigInt(0);
    let durationStr = args.shift();
    if (durationStr && /([0-9]{1,3}[smhdwy])+/g.test(durationStr)) {
        let pieces = durationStr.match(/([0-9]+[smhdwy])/g) ?? [];

        // Being able to specify the same letter multiple times
        // (e.g. 1s1s) and having their values stack is a feature
        for (const piece of pieces) {
            let [num, letter] = [
                BigInt(piece.slice(0, piece.length - 1)),
                piece.slice(piece.length - 1),
            ];
            let multiplier = 0;

            switch (letter) {
                case "s":
                    multiplier = 1000;
                    break;
                case "m":
                    multiplier = 1000 * 60;
                    break;
                case "h":
                    multiplier = 1000 * 60 * 60;
                    break;
                case "d":
                    multiplier = 1000 * 60 * 60 * 24;
                    break;
                case "w":
                    multiplier = 1000 * 60 * 60 * 24 * 7;
                    break;
                case "y":
                    multiplier = 1000 * 60 * 60 * 24 * 365;
                    break;
            }

            duration += num * BigInt(multiplier);
        }
    } else if (durationStr) args.unshift(durationStr);

    if (duration > Number.MAX_SAFE_INTEGER)
        throw new Error("Duration exceeds one morbillion years");

    return { time: Number(duration), text: args.join(" ") };
}

/**
 * Input: { a: "b", "c": { "d": "e", "f": "g" } }
 * Output: { a: "b", "c.d": "e", "c.f": "g" }
 */
export function flattenObject(
    input: object | null | undefined,
    keyPrefix: string = "",
): object {
    if (!input) return {};
    const result: { [key: string]: any } = {};

    for (const [key, value] of Object.entries(input)) {
        if (typeof value == "object" && !Array.isArray(value)) {
            const r = flattenObject(
                value,
                keyPrefix ? `${keyPrefix}.${key}` : key,
            );
            for (const [k, v] of Object.entries(r)) {
                result[k] = v;
            }
        } else {
            result[keyPrefix ? `${keyPrefix}.${key}` : key] = value;
        }
    }

    return result;
}

export async function enabledInServer(
    server: OrID<Server>,
    client: AutomodClient,
) {
    // todo implement proper server white/blacklisting system
    return ["01F80118K1F2EYD9XAMCPQ0BCT"].includes(server.id);
}

/**
 * Counts all infractions in a server and returns the ID for a new infraction based on that.
 * ID is in the format of `0001`, `0002`, etc and starts at 1. (Scales past 10000).
 * Since the next available ID is automatically updated when this is executed, the returned
 * ID should always be used.
 */
export async function generateInfractionId(
    server: string | Server,
    dbs: AutomodDatabase,
): Promise<DBInfraction["_id"]["infraction"]> {
    const res = await dbs.incrementingIDs.findOneAndUpdate(
        {
            _id: server instanceof Server ? server.id : server,
        },
        {
            $inc: { infractions: 1 },
        },
        {
            upsert: true,
        },
    );

    return String((res.value?.infractions ?? 0) + 1).padStart(4, "0");
}

// Replaces newlines with ⏎
export const hideNewlines = (input: string) => input.replaceAll("\n", "⏎");

export const limitLength = (input: string, max: number) =>
    input.length > max ? input.substring(0, max - 1) + "…" : input;

// TODO switch to a more flexible permission system
export const memberIsModerator = async (
    member: ServerMember | null | undefined,
) => (member ? member.hasPermission(member.server!, "KickMembers") : false);

export const generateAuthToken = () => {
    return crypto.randomBytes(64).toString("base64");
};

export const generateOTP = () => {
    return crypto.randomBytes(3).toString("hex");
};
